import torch
import numpy as np
import pytorch_lightning as pl
from typing import List, Optional, Dict, Tuple
from ml4floods.preprocess.worldfloods import normalize
from ml4floods.models.config_setup import  AttrDict

from ml4floods.models.utils import losses, metrics
from ml4floods.models.architectures.baselines import SimpleLinear, SimpleCNN
from ml4floods.models.architectures.unets import UNet, UNet_dropout
from ml4floods.models.architectures.hrnet_seg import HighResolutionNet
from ml4floods.data.worldfloods.configs import COLORS_WORLDFLOODS, CHANNELS_CONFIGURATIONS, BANDS_S2, COLORS_WORLDFLOODS_INVLANDWATER, COLORS_WORLDFLOODS_INVCLEARCLOUD
from pytorch_lightning.loggers import WandbLogger
from ml4floods.data.utils import get_filesystem

class WorldFloodsModel(pl.LightningModule):
    """
    Model to do multiclass classification.
    It expects ground truths y (B, H, W) tensors to be encoded as: {0: invalid, 1: clear, 2:water, 3: cloud}
    The preds (model.forward(x)) will produce a tensor with shape (B, 3, H, W)
    """
    def __init__(self, model_params: AttrDict, normalized_data:bool=True):
        super().__init__()
        self.save_hyperparameters()
        h_params_dict = model_params.get('hyperparameters', {})
        self.num_class = h_params_dict.get('num_classes', 3)
        self.network = configure_architecture(h_params_dict)
        self.weight_per_class = torch.Tensor(h_params_dict.get('weight_per_class',
                                                               [1 for i in range(self.num_class)]),
                                             device=self.device)
        self.normalized_data = normalized_data

        # learning rate params
        self.lr = h_params_dict.get('lr', 1e-4)
        self.lr_decay = h_params_dict.get('lr_decay', 0.5)
        self.lr_patience = h_params_dict.get('lr_patience', 2)
        
        #label names setup
        self.label_names = h_params_dict.get('label_names', [i for i in range(self.num_class)])

    def training_step(self, batch: Dict, batch_idx) -> float:
        """
        Args:
            batch: includes
                x (torch.Tensor): (B, C, W, H), input image
                y (torch.Tensor): (B, 1, W, H) encoded as {0: invalid, 1: land, 2: water, 3: cloud}
        """
        x, y = batch['image'], batch['mask'].squeeze(1)
        logits = self.network(x)
        loss = losses.calc_loss_mask_invalid(logits, y, weight=self.weight_per_class.to(self.device))
        if (batch_idx % 100) == 0:
            self.log("loss", loss)
        
        if (batch_idx == 0) and (self.logger is not None) and isinstance(self.logger, WandbLogger):
            with torch.no_grad():
                self.log_images(x, y, logits,prefix="train_")
            
        return loss
    
    def forward(self, x:torch.Tensor) -> torch.Tensor:
        """

        Args:
            x: (B, num_channels, H, W) input tensor

        Returns:
            (B, 3, H, W) prediction of the network
        """
        return self.network(x)

    def image_to_logger(self, x:torch.Tensor) -> Optional[np.ndarray]:
        return batch_to_unnorm_rgb(x,
                                   self.hparams["model_params"]["hyperparameters"]['channel_configuration'],
                                   unnormalize=self.normalized_data)


    def log_images(self, x, y, logits, prefix=""):
        import wandb
        mask_data = y.cpu().numpy()
        pred_data = torch.argmax(logits, dim=1).long().cpu().numpy()

        img_data = self.image_to_logger(x)

        if img_data is not None:
            self.logger.experiment.log(
                {f"{prefix}overlay": [self.wb_mask(img, pred, mask) for (img, pred, mask) in zip(img_data, pred_data, mask_data)]})
            self.logger.experiment.log({f"{prefix}image": [wandb.Image(img) for img in img_data]})

        self.logger.experiment.log({f"{prefix}y": [wandb.Image(mask_to_rgb(img)) for img in mask_data]})
        self.logger.experiment.log({f"{prefix}pred": [wandb.Image(mask_to_rgb(img + 1)) for img in pred_data]})

    def validation_step(self, batch: Dict, batch_idx):
        """
        Args:
            batch: includes
                x (torch.Tensor): (B, C, W, H), input image
                y (torch.Tensor): (B, W, H) encoded as {0: invalid, 1: land, 2: water, 3: cloud}
        """
        x, y = batch['image'], batch['mask'].squeeze(1)
        logits = self.network(x)
        
        bce_loss = losses.cross_entropy_loss_mask_invalid(logits, y, weight=self.weight_per_class.to(self.device))
        dice_loss = losses.dice_loss_mask_invalid(logits, y)
        self.log('val_bce_loss', bce_loss)
        self.log('val_dice_loss', dice_loss)

        pred_categorical = torch.argmax(logits, dim=1).long()

        # cm_batch is (B, num_class, num_class)
        cm_batch = metrics.compute_confusions(y, pred_categorical, num_class=self.num_class,
                                              remove_class_zero=True)

        # Log accuracy per class
        recall = metrics.calculate_recall(cm_batch, self.label_names)
        for k in recall.keys():
            self.log(f"val_recall {k}", recall[k])

        # Log IoU per class
        iou_dict = metrics.calculate_iou(cm_batch, self.label_names)
        for k in iou_dict.keys():
            self.log(f"val_iou {k}", iou_dict[k])
            
        if (batch_idx == 0) and (self.logger is not None) and isinstance(self.logger, WandbLogger):
            self.log_images(x, y, logits,prefix="val_")
            

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.network.parameters(), self.lr)
        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer,
                                                               mode=METRIC_MODE[self.hparams["model_params"]["hyperparameters"]["metric_monitor"]],
                                                               factor=self.lr_decay, verbose=True,
                                                               patience=self.lr_patience)

        return {"optimizer": optimizer, "lr_scheduler": scheduler,
                "monitor": self.hparams["model_params"]["hyperparameters"]["metric_monitor"]}

    def wb_mask(self, bg_img, pred_mask, true_mask):
        import wandb
        return wandb.Image(bg_img, masks={
            "prediction" : {"mask_data" : pred_mask, "class_labels" : self.labels()},
            "ground truth" : {"mask_data" : true_mask, "class_labels" : self.labels()}})
        
    def labels(self):
        return {
            0: "invalid",
            1: "land",
            2: "water",
            3: "cloud"
        }


METRIC_MODE = {
    "val_dice_loss": "min",
    "val_iou water": "max",
    "val_bce_land_water": "min",
    "val_bce_loss": "min",
    "val_Acc_land_water": "max"
}


class ML4FloodsModel(pl.LightningModule):
    """
    Model to do multioutput binary classification.
    It expects ground truths y (B, 2, H, W) tensors to be encoded as:
    - Channel 0: {0: invalid, 1: clear, 2: cloud}
    - Channel 1: {0: invalid, 1: land, 2: water}
    """
    def __init__(self, model_params: AttrDict, normalized_data:bool = True):
        super().__init__()
        self.save_hyperparameters()
        h_params_dict = model_params.get('hyperparameters', {})
        self.num_class = h_params_dict.get('num_classes', 2)
        assert self.num_class == 2, "Expected 2 output classes"

        self.pos_weight = h_params_dict.get('pos_weight',
                                            [1 for i in range(self.num_class)])
        self.weight_problem = [1 / self.num_class for _ in range(self.num_class)]

        self.network = configure_architecture(h_params_dict)
        self.normalized_data = normalized_data

        # learning rate params
        self.lr = h_params_dict.get('lr', 1e-4)
        self.lr_decay = h_params_dict.get('lr_decay', 0.5)
        self.lr_patience = h_params_dict.get('lr_patience', 2)

        # label names setup
        self.label_names = np.array(h_params_dict['label_names'])
        assert self.label_names.shape == (2, 3), "Unexpected label names, expected: {}".format([["invalid","clear", "cloud"],
                                                                                                ["invalid", "land", "water"]])

        self.colormaps = {0 : COLORS_WORLDFLOODS_INVCLEARCLOUD, 1: COLORS_WORLDFLOODS_INVLANDWATER}


    def training_step(self, batch: Dict, batch_idx) -> float:
        """
        Args:
            batch: includes
                x (torch.Tensor): (B, 2, W, H), input image
                y (torch.Tensor): (B, 2, W, H) encoded as {0: invalid, 1: neg_xxx, 2: pos_xxx}
        """
        x, y = batch['image'], batch['mask']
        logits = self.network(x)
        loss = losses.calc_loss_multioutput_logistic_mask_invalid(logits, y,
                                                                  pos_weight_problem=self.pos_weight,
                                                                  weight_problem=self.weight_problem)

        if (batch_idx % 100) == 0:
            self.log("loss", loss)

        if (batch_idx == 0) and (self.logger is not None) and isinstance(self.logger, WandbLogger):
            with torch.no_grad():
                self.log_images(x, y, logits, prefix="train_")

        return loss

    def forward(self, x):
        """

        Args:
            x: (B, num_channels, H, W) input tensor

        Returns:
            (B, 2, H, W) prediction of the network:
            - channel 0: probability of pixel being cloud
            - channel 1: probability of pixel being cwater

        """
        return self.network(x)

    def image_to_logger(self, x:torch.Tensor) -> Optional[np.ndarray]:
        return batch_to_unnorm_rgb(x,
                                   self.hparams["model_params"]["hyperparameters"]['channel_configuration'],
                                   unnormalize=self.normalized_data)

    def log_images(self, x, y, logits, prefix=""):
        import wandb
        """ Log batch images and preds using wandb """
        mask_data = y.cpu().numpy()
        pred_data = torch.round(torch.sigmoid(logits)).long().cpu().numpy()
        img_data = self.image_to_logger(x)

        self.logger.experiment.log({f"{prefix}image": [wandb.Image(img) for img in img_data]})

        for i in range(self.num_class):
            problem_name = "_".join(self.label_names[i, 1:])
            self.logger.experiment.log({f"{prefix}{problem_name}_pred_cont": [wandb.Image(img[i], mode="L") for img in pred_data]})
            self.logger.experiment.log({f"{prefix}{problem_name}_pred": [wandb.Image(mask_to_rgb(img[i].round().astype(np.int64) + 1,
                                                                                                 values=[0, 1, 2], colors_cmap=self.colormaps[i])) for img in pred_data]})
            self.logger.experiment.log({f"{prefix}y_{problem_name}": [wandb.Image(mask_to_rgb(img[i], values=[0, 1, 2], colors_cmap=self.colormaps[i])) for img in mask_data]})

    def validation_step(self, batch: Dict, batch_idx):
        """
        Args:
            batch: includes
                x (torch.Tensor): (B, C, W, H), input image
                y (torch.Tensor): (B, W, H) encoded as {0: invalid, 1: land, 2: water, 3: cloud}
        """
        x, y = batch['image'], batch['mask']
        logits = self.network(x)

        bce_loss = losses.calc_loss_multioutput_logistic_mask_invalid(logits, y)
        self.log('val_bce_loss', bce_loss)

        pred_categorical = torch.round(torch.sigmoid(logits)).long()

        # cm_batch is (B, num_class, num_class)

        for i in range(len(self.label_names)):
            cm_batch = metrics.compute_confusions(y[:, i], pred_categorical[:, i],
                                                  num_class=2, remove_class_zero=True)

            problem_name = "_".join(self.label_names[i, 1:])

            cm_agg = torch.sum(cm_batch, dim=0)
            self.log(f"val_Acc_{problem_name}", metrics.binary_accuracy(cm_agg))
            self.log(f"val_Precision_{problem_name}", metrics.binary_precision(cm_agg))
            self.log(f"val_Recall_{problem_name}", metrics.binary_recall(cm_agg))

            bce = losses.binary_cross_entropy_loss_mask_invalid(logits[:, i], y[:, i], pos_weight=None)
            self.log(f"val_bce_{problem_name}", bce)

            # Log IoU per class
            iou_dict = metrics.calculate_iou(cm_batch, self.label_names[i, 1:])
            for k in iou_dict.keys():
                self.log(f"val_iou_{problem_name} {k}", iou_dict[k])

        if (batch_idx == 0) and (self.logger is not None) and isinstance(self.logger, WandbLogger):
            self.log_images(x, y, logits, prefix="val_")

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.network.parameters(), self.lr)
        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer,
                                                               mode=METRIC_MODE[self.hparams["model_params"]["hyperparameters"]["metric_monitor"]],
                                                               factor=self.lr_decay, verbose=True,
                                                               patience=self.lr_patience)

        return {"optimizer": optimizer, "lr_scheduler": scheduler,
                "monitor": self.hparams["model_params"]["hyperparameters"]["metric_monitor"]}


def mask_to_rgb(mask:np.ndarray, values:List[int]=[0, 1, 2, 3],
                colors_cmap:np.ndarray=COLORS_WORLDFLOODS) -> np.ndarray:
    """
     Given a 2D mask it assign each value of the mask the corresponding color
    Args:
        mask: (H, W) tensor with values in values
        values: list of values to replace by the colors in colors_cmap
        colors_cmap:

    Returns:

    """
    assert len(values) == len(
        colors_cmap), f"Values and colors should have same length {len(values)} {len(colors_cmap)}"
    assert len(mask.shape) == 2, f"Unexpected shape {mask.shape}"
    mask_return = np.zeros(mask.shape[:2] + (3,), dtype=np.uint8)
    colores = np.array(np.round(colors_cmap * 255), dtype=np.uint8)
    for i, c in enumerate(colores):
        mask_return[mask == values[i], :] = c
    return mask_return


PATH_TO_MODEL_HRNET_SMALL = "gs://ml4cc_data_lake/0_DEV/2_Mart/2_MLModelMart/hrnet_small_imagenet/HRNet_W18_C_ssld_pretrained.pth"


def configure_architecture(h_params:AttrDict) -> torch.nn.Module:
    architecture = h_params.get('model_type', 'linear')
    num_channels = h_params.get('num_channels', 3)
    num_classes = h_params.get('num_classes', 2)

    if architecture == 'unet':
        model = UNet(num_channels, num_classes)

    elif architecture == 'simplecnn':
        model = SimpleCNN(num_channels, num_classes)

    elif architecture == 'linear':
        model = SimpleLinear(num_channels, num_classes)

    elif architecture == 'unet_dropout':
        model = UNet_dropout(num_channels, num_classes)

    elif architecture == "hrnet_small":
        model = HighResolutionNet(input_channels=num_channels, output_channels=num_classes)
        if num_channels == 3:
            print("3-channel model. Loading pre-trained weights from ImageNet")
            pretrained_dict = load_weights(PATH_TO_MODEL_HRNET_SMALL)
            model.init_weights(pretrained_dict)

    else:
        raise Exception(f'No model implemented for model_type: {h_params.model_type}')

    return model


def load_weights(path_weights:str, map_location="cpu"):
    fs = get_filesystem(path_weights)
    if fs.exists(path_weights):
        with fs.open(path_weights, "rb") as fh:
            weights = torch.load(fh, map_location=map_location)

        return weights

    raise ValueError(f"Pretrained weights file: {path_weights} does not exists")


def batch_to_unnorm_rgb(x:torch.Tensor, channel_configuration:str="all", max_clip_val=3000.,
                        unnormalize:bool=True) -> Optional[np.ndarray]:
    """
    Unnorm x images and get rgb channels for visualization

    Args:
        x: (B, C, H, W) image
        channel_configuration: one of CHANNELS_CONFIGURATIONS.keys()
        max_clip_val: value to saturate the image
        unnormalize:

    Returns:
        (B, H, W, 3) np.array with values between 0-1 ready to be used in imshow/PIL.from_array()
    """
    model_input_npy = x.cpu().numpy()

    # Find the RGB indexes within the S2 bands
    bands_read_names = [BANDS_S2[i] for i in CHANNELS_CONFIGURATIONS[channel_configuration]]
    bands_index_rgb = [bands_read_names.index(b) for b in ["B4", "B3", "B2"] if b in bands_read_names]
    if len(bands_index_rgb) != 3:
        # try swir/nir/red
        bands_index_rgb = [bands_read_names.index(b) for b in ["B11", "B8", "B4"] if b in bands_read_names]
        if len(bands_index_rgb) != 3:
            return

    mean, std = normalize.get_normalisation("rgb")  # B, R, G!
    mean = mean[np.newaxis] # (1, 1, 1, nchannels)
    std = std[np.newaxis] # (1, 1, 1, nchannels)

    model_input_rgb_npy = model_input_npy[:, bands_index_rgb, ...].transpose(0, 2, 3, 1)
    if unnormalize:
        model_input_rgb_npy = model_input_rgb_npy  * std + mean
        model_input_rgb_npy = np.clip(model_input_rgb_npy / max_clip_val, 0., 1.)

    return model_input_rgb_npy


def unnorm_batch(x:torch.Tensor, channel_configuration:str="all", max_clip_val:float=3000.) ->torch.Tensor:
    model_input_npy = x.cpu().numpy()

    mean, std = normalize.get_normalisation(channel_configuration, channels_first=True)
    mean = mean[np.newaxis] # (1, nchannels, 1, 1)
    std = std[np.newaxis]  # (1, nchannels, 1, 1)
    out = model_input_npy * std + mean
    if max_clip_val is not None:
        out = np.clip(out/max_clip_val, 0, 1)
    return out


def plot_batch(x:torch.Tensor, channel_configuration:str="all", bands_show=None, axs=None, max_clip_val=3000.,
               show_axis=False):
    """

    Args:
        x:
        channel_configuration:
        bands_show: RGB ["B4", "B3", "B2"] SWIR/NIR/RED ["B11", "B8", "B4"]
        axs:
        max_clip_val: value to saturate the image
        show_axis: Whether to show axis of the image or not

    Returns:

    """
    import matplotlib.pyplot as plt

    if bands_show is None:
        bands_show = ["B4", "B3", "B2"]

    if axs is None:
        fig, axs = plt.subplots(1, len(x))

    x = unnorm_batch(x, channel_configuration, max_clip_val=max_clip_val)
    bands_read_names = [BANDS_S2[i] for i in CHANNELS_CONFIGURATIONS[channel_configuration]]
    bands_index_rgb = [bands_read_names.index(b) for b in bands_show]
    x = x[:, bands_index_rgb]

    if hasattr(x, "cpu"):
        x = x.cpu()

    for xi, ax in zip(x, axs):
        xi = np.transpose(xi, (1, 2, 0))
        ax.imshow(xi)
        if not show_axis:
            ax.axis("off")
        


def plot_batch_output_v1(outputv1: torch.Tensor, axs=None, legend=True, show_axis=False):
    """

    Args:
        outputv1:  (B, W, H) Tensor encoded as {0: invalid, 1: land, 2: water, 3: cloud}
        axs:
        legend: whether to show the legend or not
        show_axis:  Whether to show axis of the image or not

    Returns:

    """
    import matplotlib.pyplot as plt
    from ml4floods.visualization import plot_utils

    if hasattr(outputv1, "cpu"):
        outputv1 = outputv1.cpu()

    if axs is None:
        axs = plt.subplots(1, len(outputv1))

    cmap_preds, norm_preds, patches_preds = plot_utils.get_cmap_norm_colors(plot_utils.COLORS_WORLDFLOODS,
                                                                            plot_utils.INTERPRETATION_WORLDFLOODS)

    for _i, (xi, ax) in enumerate(zip(outputv1, axs)):
        ax.imshow(xi, cmap=cmap_preds, norm=norm_preds,
                  interpolation='nearest')

        if not show_axis:
            ax.axis("off")

        if _i == (len(outputv1)-1) and legend:
            ax.legend(handles=patches_preds,
                      loc='upper right')









import math
import os
import sys
from typing import List

import rasterio

from ml4floods.preprocess.tiling import WindowSize, WindowSlices, get_window_tiles


def get_list_of_window_slices(
    file_names: List[str], window_size: WindowSize
) -> List[WindowSlices]:
    """Function to return the list of window slices for the all the
    input images and the given window size.

    Args:
        file_names (List[str]): List of filenames that are to be sliced.
        window_size (WindowSize): Window size of the tiles.

    Returns:
        List[WindowSlices]: List of window slices for the each tile
        corresponding to each input image.
    """

    accumulated_list_of_windows = []
    for ifilename in file_names:

        with rasterio.open(ifilename) as dataset:
            # get list of windows
            list_of_windows = get_window_tiles(
                dataset, height=window_size.height, width=window_size.width
            )
            # create a list of filenames
            list_of_windows = [
                WindowSlices(file_name=ifilename, window=iwindow)
                for iwindow in list_of_windows
            ]

        accumulated_list_of_windows += list_of_windows

    return accumulated_list_of_windows



from typing import Dict, Tuple

import albumentations
import cv2
import numpy as np
import torch
from albumentations import (Compose, Flip, GaussNoise, MotionBlur, Normalize,
                            PadIfNeeded, RandomRotate90, ShiftScaleRotate)
from albumentations.augmentations import functional as F
from albumentations.core.composition import BaseCompose
from albumentations.core.transforms_interface import (BasicTransform,
                                                      DualTransform)

from ml4floods.preprocess.worldfloods.normalize import get_normalisation

# TODO: split the ToTensor to ToImageTensor and ToMaskTensor for better clarity.
# TODO: separate functions for the operators (e.g. permute_channels)


def permute_channels(
    image: np.ndarray,
) -> np.ndarray:
    return image.transpose(2, 0, 1)


class ToTensor(BasicTransform):
    def __init__(self):
        super(ToTensor, self).__init__(always_apply=True, p=1.0)

    # def __call__(self, input_data: dict, force_apply=True) -> dict:
    def __call__(self, force_apply=True, **data) -> Tuple[torch.Tensor, torch.Tensor]:
        # Convert image to tensor
        image, mask = data["image"], data["mask"]

        # convert to tensor
        image = self._image_to_tensor(image)
        mask = self._mask_to_tensor(mask)

        # create output dictionary
        data["image"], data["mask"] = image, mask
        return data

    def _image_to_tensor(self, image: np.ndarray) -> torch.Tensor:
        return torch.from_numpy(image)

    def _mask_to_tensor(self, mask):
        mask_tensor = torch.from_numpy(mask).long()
        return mask_tensor

    @property
    def targets(self):
        raise NotImplementedError


class PermuteChannels(BasicTransform):
    def __init__(self):
        super(PermuteChannels, self).__init__(always_apply=True, p=1.0)

    def __call__(self, force_apply=True, **data) -> Tuple[torch.Tensor, torch.Tensor]:
        # Convert image to tensor
        image, mask = data["image"], data["mask"]

        image = self._permute_channels(image)
        mask = self._permute_channels(mask)

        # create output dictionary
        data["image"], data["mask"] = image, mask
        return data

    def _permute_channels(self, input_image: np.ndarray) -> np.ndarray:
        if input_image.ndim == 3:
            input_image = input_image.transpose(2, 0, 1)
        return input_image

    @property
    def targets(self):
        raise NotImplementedError


class InversePermuteChannels(BasicTransform):
    def __init__(
        self,
    ):
        super(InversePermuteChannels, self).__init__(always_apply=True, p=1.0)

    def __call__(self, force_apply=True, **data) -> Tuple[torch.Tensor, torch.Tensor]:
        # Convert image to tensor
        image, mask = data["image"], data["mask"]

        image = self._inverse_permute_channels(image)
        mask = self._inverse_permute_channels(mask)

        # create output dictionary
        data["image"], data["mask"] = image, mask
        return data

    def _inverse_permute_channels(self, input_image: np.ndarray) -> np.ndarray:
        if input_image.ndim == 3:
            input_image = input_image.transpose(1, 2, 0)
        return input_image

    @property
    def targets(self):
        raise NotImplementedError


class OneHotEncoding(BasicTransform):
    """One hot encode the input ground truth labels.
    TODO: Make sure that we are getting the output shape as desired
        - current output shape = C x H x W x num_classes
    """

    def __init__(self, num_classes: int):
        super(OneHotEncoding, self).__init__(always_apply=True, p=1.0)
        self.num_classes = num_classes

    def __call__(self, force_apply=True, **data) -> Tuple[torch.Tensor, torch.Tensor]:
        # Convert image to tensor
        image, mask = data["image"], data["mask"]

        mask = self._one_hot_encode(mask)

        # create output dictionary
        data["image"], data["mask"] = image, mask
        return data

    def _one_hot_encode(self, input_mask: torch.Tensor) -> torch.Tensor:
        if self.num_classes > 1:
            input_mask = torch.nn.functional.one_hot(input_mask, self.num_classes)

        return input_mask

    @property
    def targets(self):
        raise NotImplementedError


# IGNORE FOR NOW
class PerChannel(BaseCompose):
    """Apply transformations per-channel

    Args:
        transforms (list): list of transformations to compose.
        channels (list): channels to apply the transform to. Pass None to apply to all.
                         Default: None (apply to all)
        p (float): probability of applying the transform. Default: 0.5.
    """

    def __init__(self, transforms, channels=None, p=0.5):
        super(PerChannel, self).__init__(transforms, p)
        self.transforms = transforms
        self.channels = channels

    def __call__(self, force_apply=True, **data) -> Tuple[torch.Tensor, torch.Tensor]:
        # Convert image to tensor
        image, mask = data["image"], data["mask"]
        print(image.shape)
        # Mono images
        if image.ndim == 2:
            image = np.expand_dims(image, 0)

        if self.channels is None:
            self.channels = range(image.shape[-1])

        for ichannel in self.channels:
            for itransform in self.transforms:
                # get result (dictionary output)

                res = itransform(image=image[:, :, ichannel])
                # augment the channel of the dictionary
                image[:, :, ichannel] = res["image"]

        # create output dictionary
        data["image"], data["mask"] = image, mask
        return data


class ResizeFactor(DualTransform):
    """Resize the input by the given factor (assume input (rows,cols,channels) or (rows,cols)

    Args:
        p (float): probability of applying the transform. Default: 1.
        downsampling_factor (int): desired height of the output.
        interpolation (OpenCV flag): flag that is used to specify the interpolation algorithm. Should be one of:
            cv2.INTER_NEAREST, cv2.INTER_LINEAR, cv2.INTER_CUBIC, cv2.INTER_AREA, cv2.INTER_LANCZOS4.
            Default: cv2.INTER_LINEAR.

    Targets:
        image, mask, bboxes

    Image types:
        uint8, float32
    """

    def __init__(
        self,
        downsampling_factor,
        interpolation=cv2.INTER_LINEAR,
        always_apply=True,
        p=1,
    ):
        super(ResizeFactor, self).__init__(always_apply, p)
        self.downsampling_factor = downsampling_factor
        self.interpolation = interpolation

    def __call__(self, force_apply=True, **data) -> Tuple[torch.Tensor, torch.Tensor]:
        image, mask = data["image"], data["mask"]

        new_size = np.round(
            np.array(image.shape[:2]) / self.downsampling_factor
        ).astype(np.int64)
        image = F.resize(
            image,
            height=new_size[0],
            width=new_size[1],
            interpolation=self.interpolation,
        )
        # create output dictionary
        data["image"], data["mask"] = image, mask
        return data


def transforms_generator(config: Dict) -> albumentations.Compose:
    """Function to create a transformation composition using the parameters of a config file.
    This motive of this function is to provide an ability to modify transformations without changing the code.

    Args:
        config (Dict): Configuration file read a dictionary

    Returns:
        albumentations.Compose: Returns the composition of all the transformations mentioned in the config file.
    """
    # initialize list of transformations
    list_of_transforms = []
    list_of_transforms += [
        InversePermuteChannels(),
    ]
    # populate arguments
    if config["resizefactor"] is not None:
        transform_resize = ResizeFactor(**config["resizefactor"])
        list_of_transforms += [
            transform_resize,
        ]

    if config["use_channels"] is not None:
        channel_mean, channel_std = get_normalisation(config["use_channels"])
        transform_normalize = Normalize(
            mean=channel_mean, std=channel_std, max_pixel_value=1.0
        )
        list_of_transforms += [
            transform_normalize,
        ]

    if config["gaussnoise"] is not None:
        transform_gaussnoise = GaussNoise(
            var_limit=(
                config["gaussnoise"]["var_limit_lower"],
                config["gaussnoise"]["var_limit_upper"],
            ),
            p=config["gaussnoise"]["p"],
        )
        list_of_transforms += [transform_gaussnoise]

    if config["motionblur"] is not None:
        transform_motionblue = MotionBlur(**config["motionblur"])
        list_of_transforms += [transform_motionblue]

    list_of_transforms += [
        PermuteChannels(),
    ]

    if config["totensor"] is True:
        list_of_transforms += [ToTensor()]

#     if config["num_classes"] > 1:
#         list_of_transforms += [OneHotEncoding(num_classes=3)]

    return albumentations.Compose(list_of_transforms)



import torch
import itertools
from ml4floods.models.worldfloods_model import WorldFloodsModel, ML4FloodsModel, load_weights
from ml4floods.models.utils.configuration import AttrDict
from ml4floods.data.worldfloods.configs import CHANNELS_CONFIGURATIONS, SENTINEL2_NORMALIZATION, CHANNELS_CONFIGURATIONS_LANDSAT, BANDS_S2, BANDS_L8
import numpy as np
import os

from typing import (Callable, Dict, Iterable, List, NamedTuple, Optional,
                    Tuple, Union)


# U-Net inputs must be divisible by 8
SUBSAMPLE_MODULE = {
    "unet": 8,
    "unet_dropout": 8
}

def get_model(model_params:AttrDict,
              experiment_name:Optional[str]=None, normalized_data:bool=True) -> Union[ML4FloodsModel, WorldFloodsModel]:
    """
    Creates a model from a model_params Dict. (i.e. config.model_params).
    if model_params is in test or deploy model it will load the weights in the model_params.model_folder

    Args:
        model_params:
        experiment_name:
        normalized_data:

    Returns:
        pytorch lightning model

    """
    if model_params.get("test", False) or model_params.get("deploy", False):
        assert experiment_name is not None, f"Expermient name must be set on test or deploy mode"

        if model_params.get("model_version", "v1") == "v2":
            model = ML4FloodsModel(model_params)
        else:
            model = WorldFloodsModel(model_params, normalized_data=normalized_data)

        path_to_models = os.path.join(model_params.model_folder, experiment_name, "model.pt").replace("\\", "/")
        model.load_state_dict(load_weights(path_to_models))
        print(f"Loaded model weights: {path_to_models}")
        return model

    elif model_params.get("train", False):

        if model_params.get("model_version", "v1") == "v2":
            return ML4FloodsModel(model_params)

        return WorldFloodsModel(model_params)

    else:
        raise Exception("No model type set in config e.g model_params.test == True")

        
        
def get_channel_configuration_bands(channel_configuration:str, collection_name:str="S2",
                                    as_string:bool=False) -> Union[List[int], List[str]]:
    """
    Returns 0-based list of channels of a given configuration name
    """
    if collection_name == "S2":
        channel_indexes = CHANNELS_CONFIGURATIONS[channel_configuration]
        if as_string:
            return [BANDS_S2[idx] for idx in channel_indexes]
        return channel_indexes
    elif collection_name =="Landsat":
        channel_indexes = CHANNELS_CONFIGURATIONS_LANDSAT[channel_configuration]
        if as_string:
            return [BANDS_S2[idx] for idx in channel_indexes]
        return channel_indexes
    else:
        raise NotImplementedError(f"Collection {collection_name} not implemented")

        
        
def get_model_inference_function(model: torch.nn.Module, config: AttrDict,
                                 apply_normalization:bool=True, eval_mode:bool=True,
                                 activation:Optional[str]=None,
                                 device:Optional[torch.device]=None) -> Callable[[torch.Tensor], torch.Tensor]:
    """
    Loads a model inference function for an specific configuration. It loads the model, the weights and ensure that
    prediction does not break bc of memory errors when predicting large tiles.

    Args:
        model :LightingModule or torch.nn.Module
        config: AttrDict with fields:
            - config.model_params.hyperparameters.model_type
            - config.model_params.hyperparameters.channel_configuration
            - config.model_params.model_version
            - config.model_params.max_tile_size
        apply_normalization:
        eval_mode: set for predicting model.eval()
        activation: activation function to apply on inference time (softmax|sigmoid). If None it will gess it from
            the model_version
        device: if None uses model.device

    Returns: callable function
    """
    print("Getting model inference function")
    model_type = config.model_params.hyperparameters.model_type
    module_shape = SUBSAMPLE_MODULE.get(model_type, 1)

    if apply_normalization:
        channel_configuration_bands = get_channel_configuration_bands(config.model_params.hyperparameters.channel_configuration)

        mean_batch = SENTINEL2_NORMALIZATION[channel_configuration_bands, 0]
        mean_batch = torch.tensor(mean_batch[None, :, None, None])  # (1, num_channels, 1, 1)

        std_batch = SENTINEL2_NORMALIZATION[channel_configuration_bands, 1]
        std_batch = torch.tensor(std_batch[None, :, None, None])  # (1, num_channels, 1, 1)

        def normalize(batch_image):
            assert batch_image.ndim == 4, "Expected 4d tensor"
            return (batch_image - mean_batch) / (std_batch + 1e-6)
    else:
        normalize = None

    # Guess activation function
    if activation is None:
        if model_type.endswith("ndwi"):
            activation = "None"
        elif config.model_params.get("model_version", "v1") == "v2":
            activation = "sigmoid"
        else:
            activation = "softmax"

    if activation == "None":
        activation_fun = lambda ot: ot
    elif activation == "softmax":
        activation_fun = lambda ot: torch.softmax(ot, dim=1)
    elif activation == "sigmoid":
        activation_fun = lambda ot: torch.sigmoid(ot)
    else:
        raise NotImplementedError(f"Activation function {activation} not implemented")

    device = device if device is not None else model.device

    return get_pred_function(model, device,
                             module_shape=module_shape,
                             max_tile_size=config.model_params.max_tile_size,
                             activation_fun=activation_fun,
                             normalization=normalize, eval_mode=eval_mode)


def get_pred_function(model: torch.nn.Module, device:torch.device, module_shape: int=1, max_tile_size: int=128,
                      normalization: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
                      activation_fun: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
                      eval_mode:bool=True) -> Callable[[torch.Tensor], torch.Tensor]:
    """
    Given a model it returns a callable function to make inferences that:
    1) Normalize the input tensor if provided a callable normalization fun
    2) Tile the input if it's bigger than max_tile_size to avoid memory errors (see pred_by_tile fun)
    3) Checks the input to the network is divisible by module_shape and pads if needed
    (to avoid errors in U-Net like models)
    4) Applies activation function to the outputs

    Args:
        model:
        device:
        module_shape:
        max_tile_size:
        normalization:
        activation_fun:
        eval_mode:

    Returns:
        Function to make inferences:  (B, C, H, W) -> (B, C, H, W)

    """
    if eval_mode:
        model.eval()
    else:
        model.train()
    
    if normalization is None:
        normalization = lambda ti: ti

    if activation_fun is None:
        activation_fun = lambda ot: ot
    
    # Pad the input to be divisible by module_shape (otherwise U-Net model fails)
    if module_shape > 1:
        pred_fun = padded_predict(lambda ti: activation_fun(model(ti.to(device))),
                                  module_shape=module_shape)
    else:
        pred_fun = lambda ti: activation_fun(model(ti.to(device)))

    def pred_fun_final(ti):
        with torch.no_grad():
            ti_norm = normalization(ti)
            if any((s > max_tile_size for s in ti.shape[2:])):
                return predbytiles(pred_fun,
                                   input_batch=ti_norm,
                                   tile_size=max_tile_size)
            
            return pred_fun(ti_norm)

    return pred_fun_final


def padded_predict(predfunction: Callable, module_shape: int) -> Callable:
    """
    This function is needed for U-Net like models that require the shape to be multiple of 8 (otherwise there is an
    error in concat layer between the tensors of the upsampling and downsampling paths).

    Args:
        predfunction:
        module_shape:

    Returns:
        Function that pads the input if it is not multiple of module_shape

    """
    def predict(x: torch.Tensor):
        """

        Args:
            x:

        Returns:

        """
        shape_tensor = np.array(list(x.shape))[2:].astype(np.int64)
        shape_new_tensor = np.ceil(shape_tensor.astype(np.float32) / module_shape).astype(np.int64) * module_shape

        if np.all(shape_tensor == shape_new_tensor):
            return predfunction(x)

        pad_to_add = shape_new_tensor - shape_tensor
        refl_pad_layer = torch.nn.ReflectionPad2d((0, pad_to_add[1], 0, pad_to_add[0]))

        refl_pad_result = refl_pad_layer(x)
        pred_padded = predfunction(refl_pad_result)
        slice_ = (slice(None),
                  slice(None),
                  slice(0, shape_new_tensor[0]-pad_to_add[0]),
                  slice(0, shape_new_tensor[1]-pad_to_add[1]))

        return pred_padded[slice_]

    return predict


def predbytiles(pred_function: Callable[[torch.Tensor], torch.Tensor], input_batch: torch.Tensor,
                tile_size=1280, pad_size=32, device=torch.device("cpu")) -> torch.Tensor:
    """
    Apply a pred_function (usually a torch model) by tiling the input_batch array.
    The purpose is to run `pred_function(input_batch)` avoiding memory errors.
    It tiles and stiches the pateches with padding using the strategy of: https://arxiv.org/abs/1805.12219

    Args:
        pred_function: pred_function to call
        input_batch: torch.Tensor in BCHW format
        tile_size: Size of the tiles.
        pad_size: each tile is padded before calling the pred_function.
        device: Device to save the predictions

    Returns:
        torch.Tensor in BCHW format (same B, H and W as input_batch)

    """
    pred_continuous_tf = None
    assert input_batch.dim() == 4, "Expected batch of images"

    for b, i, j in itertools.product(range(0, input_batch.shape[0]),
                                     range(0, input_batch.shape[2], tile_size),
                                     range(0, input_batch.shape[3], tile_size)):

        slice_current = (slice(i, min(i + tile_size, input_batch.shape[2])),
                         slice(j, min(j + tile_size, input_batch.shape[3])))
        slice_pad = (slice(max(i - pad_size, 0), min(i + tile_size + pad_size, input_batch.shape[2])),
                     slice(max(j - pad_size, 0), min(j + tile_size + pad_size, input_batch.shape[3])))

        slice_save_i = slice(slice_current[0].start - slice_pad[0].start,
                             None if (slice_current[0].stop - slice_pad[0].stop) == 0 else slice_current[0].stop -
                                                                                           slice_pad[0].stop)
        slice_save_j = slice(slice_current[1].start - slice_pad[1].start,
                             None if (slice_current[1].stop - slice_pad[1].stop) == 0 else slice_current[1].stop -
                                                                                           slice_pad[1].stop)

        slice_save = (slice_save_i, slice_save_j)

        slice_prepend = (slice(b, b + 1), slice(None))
        slice_current = slice_prepend + slice_current
        slice_pad = slice_prepend + slice_pad
        slice_save = slice_prepend + slice_save

        vals_to_predict = input_batch[slice_pad]
        cnn_out = pred_function(vals_to_predict).to(device)

        assert cnn_out.dim() == 4, "Expected 4-band prediction (after softmax)"

        if pred_continuous_tf is None:
            pred_continuous_tf = torch.zeros((input_batch.shape[0], cnn_out.shape[1],
                                              input_batch.shape[2], input_batch.shape[3]),
                                             device=device)

        pred_continuous_tf[slice_current] = cnn_out[slice_save]

    return pred_continuous_tf


import torch

# s2bands (0-based index):
#     green = 2
#     nir = 7
#     swir_cirrus = 10
#     swir11 = 11
#     swir12 = 12


def extract_ndwi(x: torch.Tensor, green:int=2, nir:int=7, epsilon:float=1e-8) -> torch.Tensor:
    """
    Normalize difference water index (Mcfeeters 1996) Band index defaults to Sentinel-2 image bands.

    Args:
        x: BCHW image
        green: index of green band
        nir: index of nir band
        epsilon: eps value to avoid dividing by zero

    Returns:
        B1HW image with the ndwi
    """
    band_sum = x[:, green, :, :] + x[:, nir, :, :]
    band_diff = x[:, green, :, :] - x[:, nir, :, :]
    ndwi = band_diff / (band_sum + epsilon)

    return ndwi[:, None]

def extract_mndwi(x: torch.Tensor, green:int=2, swir:int=11, epsilon:float=1e-8) -> torch.Tensor:
    """
    Modified normalize difference water index (Xu 2006) Band index defaults to Sentinel-2 image bands.

    Args:
        x: BCHW image
        green: index of green band
        swir: index of swir band (if Sentinel-2 could be indexes 11 or 12)
        epsilon: eps value to avoid dividing by zero

    Returns:
        B1HW image with the nndwi
    """
    band_sum = x[:, green, :, :] + x[:, swir, :, :]
    band_diff = x[:, green, :, :] - x[:, swir, :, :]
    ndwi = band_diff / (band_sum + epsilon)

    return ndwi[:, None]




"""
Part of the code in this module is adapted from https://github.com/kappazeta/cm_predict

Copyright 2020 KappaZeta Ltd.

Licensed under the Apache License, Version 2.0 (the 'License');
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0

"""
import tensorflow as tf
import os
import numpy as np
from typing import Callable, Tuple, Optional
import itertools
from ml4floods.visualization.plot_utils import get_cmap_norm_colors
from ml4floods.models import postprocess
import geopandas as gpd
import pandas as pd
import rasterio.transform


COLORS_KAPPAZETA = np.array([[255, 0, 0], # 0: invalid
                             [139, 64, 0], # 1: land
                             [20, 20, 20], # 2: shadow
                             [200, 200, 200], # 3: thin cloud
                             [240, 240, 240], # 3: thick cloud
                             [255, 0, 0]], # 5: MISSING
                            dtype=np.float32) / 255


CLASSES_KAPPAZETA = ["UNDEFINED", "CLEAR", "CLOUD_SHADOW", "SEMI_TRANSPARENT_CLOUD", "CLOUD", "MISSING"]


def plot_pred(pred_argmax:np.ndarray, ax=None, legend=True):
    """
    Plot kappazeta cloud mask in matplotlib

    Args:
        pred_argmax: image in format (H, W)
        classes:
        colors:
        ax:
        legend:

    Returns:

    """

    colors = COLORS_KAPPAZETA
    classes = CLASSES_KAPPAZETA

    cmap_preds, norm_preds, patches_preds = get_cmap_norm_colors(colors, classes)

    if ax is None:
        import matplotlib.pyplot as plt
        ax = plt.gca()
    ax.imshow(pred_argmax, cmap=cmap_preds, norm=norm_preds, interpolation='nearest')

    if legend:
        ax.legend(handles=patches_preds,
                  loc='upper right')
    return ax


def vectorize_output(prediction:np.ndarray, crs:str, transform:rasterio.transform.Affine) -> Optional[gpd.GeoDataFrame]:
    """
    Vectorize cloud and shadow classes

    Args:
        prediction: (H, W) array with predictions. Values expected in range(len(CLASSES_KAPPAZETA))
        crs:
        transform:

    Returns:
        gpd.GeoDataFrame with vectorized cloud, shadows and thick and thin clouds classes
    """
    data_out = []
    start = 0
    for c in [2, 3, 4]:
        if c == 3:
            binary_mask = (prediction == 3) | (prediction == 4)
            class_name = "THICK AND THIN CLOUDS"
        else:
            binary_mask = prediction == c
            class_name = CLASSES_KAPPAZETA[c]
        geoms_polygons = postprocess.get_water_polygons(binary_mask,
                                                        transform=transform)
        if len(geoms_polygons) > 0:
            data_out.append(gpd.GeoDataFrame({"geometry": geoms_polygons,
                                              "id": np.arange(start, start + len(geoms_polygons)),
                                              "class": class_name},
                                             crs=crs))
            start += len(geoms_polygons)

    if len(data_out) == 1:
        return data_out[0]
    elif len(data_out) > 1:
        return pd.concat(data_out, ignore_index=True)

    return None


def predbytiles(pred_function: Callable, input_batch: np.ndarray,
                tile_size=1280, pad_size=32) -> np.ndarray:
    """
    Apply a pred_function by tiling the input_batch array.
    The purpose is to run `pred_function(input_batch)` avoiding memory errors.
    It tiles and stiches the patches with padding using the strategy described in: https://arxiv.org/abs/1805.12219

    Args:
        pred_function: pred_function to call
        input_batch: torch.Tensor in (H, W, C) format (tensorflow format!)
        tile_size: Size of the tiles.
        pad_size: each tile is padded before calling the pred_function.
        device: Device to save the predictions

    Returns:
        tensor (H, W, C) format (H and W as input_batch)

    """
    pred_continuous_tf = None
    assert input_batch.ndim == 3, f"Expected batch of images found {input_batch.shape}"

    for i, j in itertools.product(range(0, input_batch.shape[0], tile_size),
                                  range(0, input_batch.shape[1], tile_size)):

        slice_current = (slice(i, min(i + tile_size, input_batch.shape[0])),
                         slice(j, min(j + tile_size, input_batch.shape[1])))
        slice_pad = (slice(max(i - pad_size, 0), min(i + tile_size + pad_size, input_batch.shape[0])),
                     slice(max(j - pad_size, 0), min(j + tile_size + pad_size, input_batch.shape[1])))

        slice_save_i = slice(slice_current[0].start - slice_pad[0].start,
                             None if (slice_current[0].stop - slice_pad[0].stop) == 0 else slice_current[0].stop -
                                                                                           slice_pad[0].stop)
        slice_save_j = slice(slice_current[1].start - slice_pad[1].start,
                             None if (slice_current[1].stop - slice_pad[1].stop) == 0 else slice_current[1].stop -
                                                                                           slice_pad[1].stop)

        slice_save = (slice_save_i, slice_save_j)

        slice_current = slice_current + (slice(None),)
        slice_pad = slice_pad + (slice(None),)
        slice_save = slice_save + (slice(None),)

        vals_to_predict = input_batch[slice_pad]
        cnn_out = pred_function(vals_to_predict)

        assert cnn_out.ndim == 3, f"Expected 3-band prediction (after softmax) found {cnn_out.shape}"

        if pred_continuous_tf is None:
            pred_continuous_tf = np.zeros((input_batch.shape[0], input_batch.shape[1],
                                           cnn_out.shape[-1]),
                                          dtype=np.float32)

        pred_continuous_tf[slice_current] = cnn_out[slice_save]

    return pred_continuous_tf


def find_padding(v:int, divisor=16) -> Tuple[int, int]:
    v_divisible = max(divisor, int(divisor * np.ceil(v / divisor)))
    total_pad = v_divisible - v
    pad_1 = total_pad // 2
    pad_2 = total_pad - pad_1
    return pad_1, pad_2


class Unet:
    """
    Unet
    """

    def __init__(self, version:str="L1C"):
        assert version == "L1C", f"Version {version} not implemented (L2A not implemented yet)"
        self.features = ["B1", "B2", "B3", "B4", "B5", "B6", "B8", "B8A", "B9", "B10", "B11", "B12"]
        self.classes = CLASSES_KAPPAZETA
        self.max_v = np.array([0.19, 0.298, 0.279, 0.293, 0.243, 0.247, 0.317, 0.354, 0.153, 0.055, 0.24, 0.23])
        self.norm_factor = 65_535 # 2**16

    def construct(self, width=None, height=None, pretrained_weights:str=None):
        """
        Construct the model.
        """
        # For symmetrical neighbourhood, width and height must be odd numbers.
        input_shape = (width, height, len(self.features))

        with tf.name_scope("Model"):
            inputs = tf.keras.layers.Input(input_shape, name='input')

            conv1 = tf.keras.layers.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_normal')(inputs)
            conv1 = tf.keras.layers.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv1)
            pool1 = tf.keras.layers.MaxPool2D(pool_size=(2, 2))(conv1)

            conv2 = tf.keras.layers.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_normal')(pool1)
            conv2 = tf.keras.layers.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv2)
            pool2 = tf.keras.layers.MaxPool2D(pool_size=(2, 2))(conv2)

            conv3 = tf.keras.layers.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_normal')(pool2)
            conv3 = tf.keras.layers.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv3)
            pool3 = tf.keras.layers.MaxPool2D(pool_size=(2, 2))(conv3)

            conv4 = tf.keras.layers.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_normal')(pool3)
            conv4 = tf.keras.layers.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv4)
            drop4 = tf.keras.layers.Dropout(0.5)(conv4)
            pool4 = tf.keras.layers.MaxPool2D(pool_size=(2, 2))(drop4)

            conv5 = tf.keras.layers.Conv2D(1024, 3, activation='relu', padding='same', kernel_initializer='he_normal')(pool4)
            conv5 = tf.keras.layers.Conv2D(1024, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv5)
            drop5 = tf.keras.layers.Dropout(0.5)(conv5)

            up6 = tf.keras.layers.Conv2D(512, 2, activation='relu', padding='same', kernel_initializer='he_normal')(
                tf.keras.layers.UpSampling2D(size=(2, 2))(drop5))
            merge6 = tf.keras.layers.concatenate([drop4, up6], axis=3)
            conv6 = tf.keras.layers.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_normal')(merge6)
            conv6 = tf.keras.layers.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv6)

            up7 = tf.keras.layers.Conv2D(256, 2, activation='relu', padding='same', kernel_initializer='he_normal')(
                tf.keras.layers.UpSampling2D(size=(2, 2))(conv6))
            merge7 = tf.keras.layers.concatenate([conv3, up7], axis=3)
            conv7 = tf.keras.layers.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_normal')(merge7)
            conv7 = tf.keras.layers.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv7)

            up8 = tf.keras.layers.Conv2D(128, 2, activation='relu', padding='same', kernel_initializer='he_normal')(
                tf.keras.layers.UpSampling2D(size=(2, 2))(conv7))
            merge8 = tf.keras.layers.concatenate([conv2, up8], axis=3)
            conv8 = tf.keras.layers.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_normal')(merge8)
            conv8 = tf.keras.layers.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv8)

            up9 = tf.keras.layers.Conv2D(64, 2, activation='relu', padding='same', kernel_initializer='he_normal')(
                tf.keras.layers.UpSampling2D(size=(2, 2))(conv8))
            merge9 = tf.keras.layers.concatenate([conv1, up9], axis=3)

            conv9 = tf.keras.layers.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_normal')(merge9)
            conv9 = tf.keras.layers.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv9)
            conv9 = tf.keras.layers.Conv2D(2, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv9)
            conv10 = tf.keras.layers.Conv2D(len(self.classes), (1, 1), activation='softmax')(conv9)

            self.model = tf.keras.Model(inputs, conv10)

            if pretrained_weights is not None:
                self.model.load_weights(pretrained_weights)

            return self.model

    def load_weights(self, path="gs://ml4cc_data_lake/0_DEV/2_Mart/2_MLModelMart/kappazeta/l1c__091-0.42.hdf5",
                     filelocal=None):
        if path.startswith("gs"):
            import fsspec
            import tempfile

            fs = fsspec.filesystem("gs")
            remove = False
            if filelocal is None:
                with tempfile.NamedTemporaryFile(dir=".", suffix=".hdf5", prefix=os.path.basename(path),
                                                 delete=True) as fileobj:
                    filelocal = fileobj.name
                remove=True

            if not os.path.exists(filelocal):
                print(f"Downloading file to {filelocal}")
                fs.get_file(path, filelocal)

            self.model.load_weights(filelocal)

            if remove and os.path.exists(filelocal):
                os.remove(filelocal)

        elif not os.path.exists(path):
            raise FileNotFoundError(f"File {path} does not exists")
        else:
            self.model.load_weights(path)

    def predict(self, input:np.ndarray, tile_size:int=1280) -> np.ndarray:
        """

        Args:
            input: (C, H, W) tensor expected to have len(self.features) channels
            tile_size: tile size to predict

        Returns:
            (H, W) with the most likely class among the self.classes classes
        """
        assert input.ndim == 3, f"Expected 3 dims in format CHW found {input.shape}"
        assert input.shape[0] == len(self.features), f"Input tensor shape {input.shape} unexpected\n" \
                                                     f"Expected format CHW with {len(self.features)} channels corresponding to S2 bands: {self.features}"

        data_to_pred = input / self.norm_factor / self.max_v[:, None, None]

        pad_r = find_padding(data_to_pred.shape[1])
        pad_c = find_padding(data_to_pred.shape[2])
        data_to_pred = np.pad(
            data_to_pred, ((0, 0), (pad_r[0], pad_r[1]), (pad_c[0], pad_c[1])), "reflect"
        )
        data_to_pred_tf = np.transpose(data_to_pred, (1, 2, 0))

        if (data_to_pred_tf.shape[0] > tile_size) or (data_to_pred_tf.shape[1] > tile_size):
            classes_prob = predbytiles(lambda patch: self.model.predict(patch[None])[0],
                                       data_to_pred_tf)
        else:
            classes_prob = self.model.predict(data_to_pred_tf[None])[0]

        slice_rows = slice(pad_r[0], None if pad_r[1] <= 0 else -pad_r[1])
        slice_cols = slice(pad_c[0], None if pad_c[1] <= 0 else -pad_c[1])
        classes_prob = classes_prob[(slice_rows, slice_cols)]
        return np.argmax(classes_prob, axis=2).astype(np.uint8)




"""
This script contains all the utility functions that are not specific to a particular kind of dataset.
These are mainly used for explorations, testing, and demonstrations.
"""
import os
import json
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union
import pandas as pd
import warnings

import geopandas as gpd
import numpy as np
from shapely.geometry import Polygon, mapping
from shapely.ops import unary_union
from datetime import datetime

from ml4floods.data.config import CLASS_LAND_COPERNICUSEMSHYDRO
import pickle
import fsspec
from contextlib import contextmanager
import rasterio


def filter_land(gpddats: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """ Filter land from pandas dataframe (land class specified in hydrology maps from CopernicusEMS) """
    isnot_land = gpddats.obj_type.apply(
        lambda g: g not in CLASS_LAND_COPERNICUSEMSHYDRO
    )

    if np.sum(isnot_land) == gpddats.shape[0]:
        return gpddats

    gpddats_notland = gpddats[isnot_land].copy()
    if gpddats_notland.shape[0] == 0:
        return gpddats_notland

    land_geometries = gpddats.geometry[~isnot_land]
    land_geometries = unary_union(land_geometries.tolist())

    # Not all polygons are valid: filter to valid gpddats_notland
    gpddats_notland_valid = gpddats_notland[gpddats_notland["geometry"].is_valid]

    gpddats_notland_valid["geometry"] = gpddats_notland_valid.geometry.apply(
        lambda g: g.difference(land_geometries)
    )

    return gpddats_notland_valid


def filter_pols(gpddats: gpd.GeoDataFrame, pol_shapely: Polygon) -> gpd.GeoDataFrame:
    """ filter pols that do not intersects pol_shapely """
    gpddats_cp = gpddats[~(gpddats.geometry.isna() | gpddats.geometry.is_empty)].copy()

    return (
        gpddats_cp[gpddats_cp.geometry.apply(lambda g: g.intersects(pol_shapely))]
        .reset_index()
        .copy()
    )

class CustomJSONEncoder(json.JSONEncoder):

    def default(self, obj_to_encode):
        """Pandas and Numpy have some specific types that we want to ensure
        are coerced to Python types, for JSON generation purposes. This attempts
        to do so where applicable.
        """
        # Pandas dataframes have a to_json() method, so we'll check for that and
        # return it if so.
        if hasattr(obj_to_encode, "to_json"):
            return obj_to_encode.to_json()
        # Numpy objects report themselves oddly in error logs, but this generic
        # type mostly captures what we're after.
        if isinstance(obj_to_encode, np.generic):
            return obj_to_encode.item()
        # ndarray -> list, pretty straightforward.
        if isinstance(obj_to_encode, np.ndarray):
            return obj_to_encode.tolist()
        if isinstance(obj_to_encode, Polygon):
            return mapping(obj_to_encode)
        if isinstance(obj_to_encode, pd.Timestamp):
            return obj_to_encode.isoformat()
        if isinstance(obj_to_encode, datetime):
            return obj_to_encode.isoformat()
        # torch or tensorflow -> list, pretty straightforward.
        if hasattr(obj_to_encode, "numpy"):
            return obj_to_encode.numpy().tolist()
        # If none of the above apply, we'll default back to the standard JSON encoding
        # routines and let it work normally.
        return super().default(obj_to_encode)


def get_filesystem(path: Union[str, Path]):
    path = str(path)
    if "://" in path:
        # use the fileystem from the protocol specified
        return fsspec.filesystem(path.split(":", 1)[0],requester_pays = True)
    else:
        # use local filesystem
        return fsspec.filesystem("file")


def write_geojson_to_gcp(gs_path: str, geojson_val: gpd.GeoDataFrame) -> None:
    fs = get_filesystem(gs_path)
    if geojson_val.shape[0] == 0:
        warnings.warn(f"Dataframe is empty. Saving in {gs_path}")
        with fs.open(gs_path, "w") as fh:
            json.dump(eval(geojson_val.to_json()), fh)
    else:
        with fs.open(gs_path, "wb") as fh:
            geojson_val.to_file(fh, driver="GeoJSON")


def check_requester_pays_gcp_available() -> bool:
    """
    Requester pays for GCP is available from GDAL 3.4

    rasterio issue: https://github.com/rasterio/rasterio/issues/1948
    Commit GDAL: https://github.com/OSGeo/gdal/pull/3883/commits/8d551953dea9290b21bd9747ec9ed22c81ca0409

    Returns:
        True if version >= 3.4

    """
    try:
        version_gdal = rasterio.__gdal_version__
        vnumber, vsubnumber = version_gdal.split(".")[:2]
        vnumber, vsubnumber = int(vnumber), int(vsubnumber)
        if (vnumber*100+vsubnumber) >= (3*100+4):
            return True
    except Exception as e:
        pass

    return False


REQUESTER_PAYS_AVAILABLE = check_requester_pays_gcp_available()


@contextmanager
def rasterio_open_read(tifffile:str, requester_pays:bool=True) -> rasterio.DatasetReader:
    if requester_pays and tifffile.startswith("gs"):
        if REQUESTER_PAYS_AVAILABLE:
            assert "GS_USER_PROJECT" in os.environ, \
                "'GS_USER_PROJECT' env variable not found and requester_pays=True set a project name to read rasters from the bucket" \
                "(i.e. -> export GS_USER_PROJECT='myprojectname')"

            with rasterio.open(tifffile) as src:
                yield src
        else:
            fs = fsspec.filesystem("gs", requester_pays=True)
            with fs.open(tifffile, "rb") as fh:
                with rasterio.io.MemoryFile(fh.read()) as mem:
                    yield mem.open()

    else:
        with rasterio.open(tifffile) as src:
            yield src


def read_geojson_from_gcp(gs_path: str) -> gpd.GeoDataFrame:
    fs = get_filesystem(gs_path)
    with fs.open(gs_path, "rb") as fh:
        return gpd.read_file(fh)


def write_pickle_to_gcp(gs_path: str, dict_val: dict) -> None:

    fs = get_filesystem(gs_path)
    with fs.open(gs_path, "wb") as fh:
        pickle.dump(dict_val, fh)


def read_pickle_from_gcp(gs_path:str) -> dict:
    fs = get_filesystem(gs_path)
    with fs.open(gs_path, "rb") as fh:
        my_dictionary = pickle.load(fh)

    return my_dictionary


def write_json_to_gcp(gs_path: str, dict_val: dict) -> None:
    fs = get_filesystem(gs_path)

    with fs.open(gs_path, "w") as fh:
        json.dump(dict_val, fh, cls=CustomJSONEncoder)


def read_json_from_gcp(gs_path: str) ->Dict:
    fs = get_filesystem(gs_path)
    with fs.open(gs_path, "r") as fh:
        my_dictionary = json.load(fh)

    return my_dictionary




from typing import Tuple

import numpy as np

from ml4floods.data.worldfloods.configs import (CHANNELS_CONFIGURATIONS,
                                                SENTINEL2_NORMALIZATION)


def get_normalisation(use_channels: str, channels_first:bool=False) -> Tuple[np.ndarray, np.ndarray]:
    """Normalization for the S2 datasets.

    Args:
        use_channels (str): Channels that are to be used.
        channels_first (bool): whether the axis should be first or last

    Returns:
        Tuple[np.ndarray, np.ndarray]: Returns the mean and standard deviation.

    Example:
        >>> from ml4floods.preprocess.worldfloods import get_normalization
        >>> use_channels = 'all'
        >>> mu, std = get_normalization(use_channels)    
    """    

    s2_channels = CHANNELS_CONFIGURATIONS[use_channels]

    s2_norm = SENTINEL2_NORMALIZATION[s2_channels]

    #  channel stats for now
    sentinel_means = s2_norm.copy()[:, 0]
    sentinel_std = s2_norm.copy()[:, 1]

    if channels_first:
        sentinel_means = sentinel_means[..., np.newaxis, np.newaxis]  # (nchannels, 1, 1)
        sentinel_std = sentinel_std[..., np.newaxis, np.newaxis]  # (nchannels, 1, 1)
    else:
        sentinel_means = sentinel_means[np.newaxis, np.newaxis]  # (1, 1, nchannels)
        sentinel_std = sentinel_std[np.newaxis, np.newaxis]  # (1, 1, nchannels)

    return sentinel_means, sentinel_std



from argparse import ArgumentParser


class IndexerArgParser(ArgumentParser):
    """Parses Indexer command-line options and arguments."""

    def __init__(self):
        super().__init__(prog="indexer.py", description="Index pipeline output by geography and output a .pkl file for use with MapDataFactory")

        self.add_argument(
            "--worldfloods-path",
            type=str,
            default="gs://ml4floods/",
            help="Path to worldfloods directory, required only to contain tiffimages subdirectory.",
        )
        self.add_argument(
            "--output-path",
            type=str,
            help='Path of output .pkl file.',
        )
        self.add_argument(
            "--log-level",
            type=str,
            default="INFO", 
            help="One of DEBUG, INFO, WARNING, or CRITICAL"
        )
        



import sys
import warnings
import traceback

import tqdm
import pandas as pd
from datetime import datetime
from ml4floods.data import utils
from ml4floods.data.ee_download import process_metadata
import os

from pathlib import Path
from typing import Optional, Callable, Tuple, Dict, Any
import geopandas as gpd
import fsspec
from ml4floods.data import save_cog


CLOUDPROB_PARENT_PATH = "worldfloods/tiffimages"
PERMANENT_WATER_PARENT_PATH = "worldfloods/tiffimages/PERMANENTWATERJRC"
META_FLOODMAP_PARENT_PATH = "worldfloods/tiffimages/meta"
WORLDFLOODS_V0_BUCKET = "ml4floods"


def worldfloods_extra_gcp_paths(main_path: str) -> Tuple[gpd.GeoDataFrame, Optional[str], Optional[str], Dict, str]:
    """
    Given a pickle file in "gs://ml4cc_data_lake/{prod_dev}/1_Staging/WorldFloods it returns the rest of the files to
    create the full worldfloods registry (the corresponding floodmap, cloud probability and permanent water)

    Args:
        main_path: path to .piclke file in bucket

    Returns:
        Locations of corresponding  floodmap, cloud probability,  permanent water, metadata_floodmap and S2 image

    """
    fs = fsspec.filesystem("gs", requester_pays=True)
    assert fs.exists(main_path), f"File {main_path} does not exists"

    meta_floodmap = utils.read_pickle_from_gcp(main_path)

    # path to floodmap path
    floodmap_path = main_path.replace("/flood_meta/", "/floodmap_edited/").replace(".pickle", ".geojson")
    if not fs.exists(floodmap_path):
        floodmap_path = floodmap_path.replace("/floodmap_edited/", "/floodmap/")

    assert fs.exists(floodmap_path), f"Floodmap not found in {floodmap_path}"

    path_aoi = os.path.dirname(os.path.dirname(floodmap_path))

    # open floodmap with geopandas
    floodmap = utils.read_geojson_from_gcp(floodmap_path)
    floodmap_date = meta_floodmap['satellite date']

    # create permenant water path
    permanent_water_path = os.path.join(path_aoi, "PERMANENTWATERJRC", f"{floodmap_date.year}.tif").replace("\\", "/")

    if not fs.exists(permanent_water_path):
        warnings.warn(f"Permanent water {permanent_water_path}. Will not be used")
        permanent_water_path = None

    csv_path = os.path.join(path_aoi, "S2", "s2info.csv")
    metadatas2 = process_metadata(csv_path, fs=fs)
    metadatas2 = metadatas2.set_index("names2file")

    assert any(metadatas2.s2available), f"Not available S2 files for {main_path}. {metadatas2}"

    # find corresponding s2_image_path for this image
    index, s2_date = best_s2_match(metadatas2, floodmap_date)

    assert s2_date is not None, f"Not found valid S2 files for {main_path}. {metadatas2}"

    assert (s2_date - floodmap_date).total_seconds() / (3600. * 24) < 10, \
        f"Difference between S2 date {s2_date} and floodmap date {floodmap_date} is larger than 10 days"

    # add exact metadata from the csv file
    meta_floodmap["s2_date"] = s2_date
    meta_floodmap["names2file"] = index
    meta_floodmap["cloud_probability"] = metadatas2.loc[index, "cloud_probability"]
    meta_floodmap["valids"] = metadatas2.loc[index, "valids"]

    s2_image_path = os.path.join(path_aoi, "S2", index+".tif").replace("\\", "/")

    # Add cloud_probability if exists in edited
    cm_edited = s2_image_path.replace("/S2/", "/cmedited_vec/").replace(".tif", ".geojson")
    if not fs.exists(cm_edited):
        cm_edited = s2_image_path.replace("/S2/", "/cmedited/")
        if not fs.exists(cm_edited):
            cm_edited = None

    return floodmap, cm_edited, permanent_water_path, meta_floodmap, s2_image_path


def best_s2_match(metadatas2:pd.DataFrame, floodmap_date:datetime) -> Tuple[Any, datetime]:
    """
    Return s2 date posterior to the floodmap_date

    Args:
        metadatas2:
        floodmap_date:

    Returns:

    """
    index = None
    s2_date = None
    for tup in metadatas2[metadatas2.s2available].itertuples():
        date_img = tup.datetime
        if (floodmap_date < date_img) or ((floodmap_date - date_img).total_seconds() / 3600. < 10):
            if s2_date is None:
                s2_date = date_img
                index = tup.Index
            else:
                if s2_date > date_img:
                    s2_date = date_img
                    index = tup.Index
    return index, s2_date


def worldfloods_old_gcp_paths(main_path: str) -> Tuple[gpd.GeoDataFrame, str, Optional[str], Dict, str]:
    """
    Given a S2 tiff file in the V0 WorldFloods dataset it returns the rest of the anciliary files to
    the corresponding floodmap, cloud probability and permanent water

    Args:
        main_path: S2 path in gs://ml4floods/worldfloods/public folder

    Returns:
        GCPPaths with locations of corresponding  floodmap, cloud probability,  permanent water, meta_floodmap and sentinel2 image

    Examples
        >>> s2_path_file = "gs://ml4floods/worldfloods/public/train/S2/EMSR260_09RUBIERA_GRA_v2_observed_event_a.tif"
        >>> worldfloods_old_gcp_paths(GCPPath(s2_path_file))

    (GCPPath(full_path='gs://ml4floods/worldfloods/public/train/floodmaps/EMSR260_09RUBIERA_GRA_v2_observed_event_a.shp', bucket_id='ml4floods', parent_path='worldfloods/public/train/floodmaps', file_name='EMSR260_09RUBIERA_GRA_v2_observed_event_a.shp', suffix='shp'),
     GCPPath(full_path='gs://ml4floods/worldfloods/tiffimages/cloudprob/EMSR260_09RUBIERA_GRA_v2_observed_event_a.tif', bucket_id='ml4floods', parent_path='worldfloods/tiffimages/cloudprob', file_name='EMSR260_09RUBIERA_GRA_v2_observed_event_a.tif', suffix='tif'),
     'gs://ml4floods/worldfloods/tiffimages/PERMANENTWATERJRC/EMSR260_09RUBIERA_GRA_v2_observed_event_a.tif')

    """
    assert main_path.check_if_file_exists(), f"File {main_path} does not exists"

    s2_image_path = main_path

    # path to floodmap path
    floodmap_path = s2_image_path.replace("/S2/", "/floodmaps/")
    floodmap_path = floodmap_path.replace(".tif", f".shp")

    assert floodmap_path.check_if_file_exists(), f"Floodmap not found in {floodmap_path}"

    # open floodmap with geopandas
    floodmap = gpd.read_file(floodmap_path.full_path)

    # create cloudprob path
    cloudprob_path = GCPPath(
        str(
            Path(WORLDFLOODS_V0_BUCKET)
                .joinpath(CLOUDPROB_PARENT_PATH)
                .joinpath("cloudprob_edited")
                .joinpath(s2_image_path.file_name)
        )
    )
    if not cloudprob_path.check_if_file_exists():
        cloudprob_path = GCPPath(
            str(
                Path(WORLDFLOODS_V0_BUCKET)
                    .joinpath(CLOUDPROB_PARENT_PATH)
                    .joinpath("cloudprob")
                    .joinpath(s2_image_path.file_name)
            )
        )

    assert cloudprob_path.check_if_file_exists(), f"Clouds not found in {cloudprob_path}"

    # create permenant water path
    permanent_water_path = GCPPath(
        str(
            Path(WORLDFLOODS_V0_BUCKET)
                .joinpath(PERMANENT_WATER_PARENT_PATH)
                .joinpath(s2_image_path.file_name)
        )
    )

    if not permanent_water_path.check_if_file_exists():
        warnings.warn(f"Permanent water {permanent_water_path}. Will not be used")
        permanent_water_path = None


    # path to meta_floodmap
    meta_floodmap_path = GCPPath(
        str(
            Path(WORLDFLOODS_V0_BUCKET)
                .joinpath(META_FLOODMAP_PARENT_PATH)
                .joinpath(s2_image_path.file_name.replace(".tif",".json"))
        )
    )
    assert meta_floodmap_path.check_if_file_exists(), f"Meta floodmap not found in {meta_floodmap_path}"

    meta_floodmap = utils.read_json_from_gcp(meta_floodmap_path.full_path)

    return floodmap, cloudprob_path, permanent_water_path, meta_floodmap, s2_image_path


def generate_item(main_path:str, output_path:str, file_name:str,
                  overwrite:bool=False, pbar:Optional[tqdm.tqdm]=None,
                  gt_fun:Callable=None, delete_if_error:bool=True,
                  paths_function:Callable=worldfloods_old_gcp_paths) -> bool:
    """

    Generates an "element" of the WorldFloods dataset with the expected naming convention. An "element" is a set of files
    that will be used for training the WorldFloods model. These files are copied in the `output_path` folowing the naming
    convention coded in `worldfloods_output_files` function.
     These files are:
    - shp vector floodmap. (subfolder floodmap)
    - tiff permanent water (subfolder PERMANENTWATERJRC)
    - tiff gt
    - tiff S2
    - tiff cloudprob
    - json with metadata info of the ground truth

    Args:
        main_path: Path to main object. The process will search for other relevant information to create all the
        aforementioned products.
        output_path: Folder where the item will be written. See fun worldfloods_output_files for corresponding output file naming convention.
        overwrite: if False it will not overwrite existing products in path_write folder.
        file_name: Name of the file to be saved (e.g. in S2/gt/floodmap/ data will be saved with this name and the corresponding extension .tif, .geojson)
        pbar: optional progress bar with method description.
        gt_fun: one of ml4floods.data.create_gt.generate_land_water_cloud_gt or ml4floods.data.create_gt.generate_water_cloud_binary_gt.
        This function determines how the ground truth is created from the input products.
        delete_if_error: whether to delete the generated files if an error is risen
        paths_function: function to get the paths of the files for the given main_path

    Returns:
        True if success in creating all the products

    """

    fs = fsspec.filesystem("gs", requester_pays=True)

    try:
        # Check if output products exist before reading from the bucket
        if not overwrite:
            expected_outputs = worldfloods_output_files(
                output_path, file_name, permanent_water_available=True, clouds_available=False, mkdirs=False)
            fsdest = utils.get_filesystem(expected_outputs[-1])

            must_process = False
            for e in expected_outputs:
                if e and not fsdest.exists(e):
                    must_process = True
                    break

            if not must_process:
                return True

        # Get input files and check that they all exist
        floodmap, cloudprob_path, permanent_water_path, metadata_floodmap, s2_image_path = paths_function(main_path)

        # get output files
        cloudprob_path_dest, floodmap_path_dest, gt_path_dest, meta_json_path_dest, permanent_water_image_path_dest, s2_image_path_dest = worldfloods_output_files(
            output_path, file_name, permanent_water_available=permanent_water_path is not None, clouds_available=cloudprob_path is not None, mkdirs=True)
    except Exception:
        warnings.warn(f"File {main_path} problem when computing input/output names")
        traceback.print_exc(file=sys.stdout)
        return False
    try:
        # generate gt, gt meta and copy all files to path_write
        fsdest = utils.get_filesystem(s2_image_path_dest)

        if not fsdest.exists(gt_path_dest) or not fsdest.exists(meta_json_path_dest) or overwrite:
            if pbar is not None:
                pbar.set_description(f"Generating Ground Truth {file_name}...")

            # Copy s2_image_path to local before reading?
            # If so we will need also to copy cloudprob_path and permanent_water_path

            gt, gt_meta = gt_fun(
                s2_image_path,
                floodmap,
                metadata_floodmap=metadata_floodmap,
                keep_streams=True,
                cloudprob_image_path=cloudprob_path, # Could be None!
                permanent_water_image_path=permanent_water_path,  # Could be None!
            )

            if len(gt.shape) == 2:
                gt = gt[None]

            if pbar is not None:
                pbar.set_description(f"Saving GT {file_name}...")

            if gt.shape[0] == 2:
                desc = ["invalid/clear/cloud", "invalid/land/water"]
            else:
                desc = ["invalid/land/water/cloud"]

            save_cog.save_cog(gt, gt_path_dest,
                              {"crs": gt_meta["crs"], "transform":gt_meta["transform"] ,"RESAMPLING": "NEAREST",
                               "compress": "lzw", "nodata": 0}, # In both gts 0 is nodata
                              descriptions=desc,
                              tags=gt_meta)

            # upload meta json to bucket
            if pbar is not None:
                pbar.set_description(f"Saving meta {file_name}...")

            # save meta in local json file
            gt_meta["crs"] = str(gt_meta["crs"])
            gt_meta["transform"] = [gt_meta["transform"].a, gt_meta["transform"].b, gt_meta["transform"].c,
                                    gt_meta["transform"].d, gt_meta["transform"].e, gt_meta["transform"].f]

            utils.write_json_to_gcp(meta_json_path_dest, gt_meta)

        # Copy floodmap shapefiles
        if not fsdest.exists(floodmap_path_dest) or overwrite:
            if pbar is not None:
                pbar.set_description(f"Saving floodmap {file_name}...")

            utils.write_geojson_to_gcp(floodmap_path_dest, floodmap)                
        
        # Copy S2 image
        if not fsdest.exists(s2_image_path_dest) or overwrite:
            if pbar is not None:
                pbar.set_description(f"Saving S2 image {file_name}...")
            
            _copy(s2_image_path, s2_image_path_dest, fs)
        
        # Copy cloudprob
        if cloudprob_path is not None and cloudprob_path_dest and (not fsdest.exists(cloudprob_path_dest) or overwrite):
            if pbar is not None:
                pbar.set_description(f"Saving cloud probs {file_name}...")
            
            _copy(cloudprob_path, cloudprob_path_dest, fs)
        
        # Copy permanent water
        if (permanent_water_image_path_dest is not None) and (not fsdest.exists(permanent_water_image_path_dest) or overwrite):
            if pbar is not None:
                pbar.set_description(f"Saving permanent water image {file_name}...")

            _copy(permanent_water_path, permanent_water_image_path_dest, fs)

    except Exception:
        warnings.warn(f"File input: {main_path} output S2 file: {s2_image_path_dest} problem when computing Ground truth")
        traceback.print_exc(file=sys.stdout)

        if not delete_if_error:
            return False

        fsdest = utils.get_filesystem(s2_image_path_dest)
        files_to_delete = [cloudprob_path_dest, gt_path_dest, meta_json_path_dest, permanent_water_image_path_dest,
                           s2_image_path_dest, floodmap_path_dest]
        for f in files_to_delete:
            if f and fsdest.exists(f):
                print(f"Deleting file {f}")
                fsdest.delete(f)

        return False

    return True


def _copy(file_or, file_dest, fsor):    
    if file_dest.startswith("gs"):
        fsor.copy(file_or, file_dest)
    else:
        fsor.get_file(file_or, file_dest)
    

def assert_element_consistency(output_path:Path, tiff_file_name:str, warn_permanent_water:bool=True):
    """
    Assert all elements in worldlfooods_output_files exists for the given tiff_file_name.

    Args:
        output_path: path to check e.g. Path("ml4cc_data_lake/0_DEV/2_Mart/worldfloods_v2_0/train")
        tiff_file_name: e.g. "SP6_20170502_WaterExtent_WetSoil_VillaRiva.tif"
        warn_permanent_water: warn if the permanent water layer is not found

    """
    cloudprob_path_dest, floodmap_path_dest, gt_path, meta_parent_path, permanent_water_image_path_dest, s2_image_path_dest = worldfloods_output_files(output_path, tiff_file_name)

    for p in [cloudprob_path_dest, floodmap_path_dest, gt_path, meta_parent_path, s2_image_path_dest]:
        assert p.check_if_file_exists(), f"{p.full_path} not found"

    if warn_permanent_water and not permanent_water_image_path_dest.check_if_file_exists():
        warnings.warn(f"{permanent_water_image_path_dest.full_path} not found")


def worldfloods_output_files(output_path:str, file_name:str,
                             permanent_water_available:bool=True,
                             clouds_available:bool=True,
                             mkdirs:bool=False) -> Tuple[str, str, str, str, Optional[str], str]:
    """
    For a given file (`tiff_file_name`) it returns the set of paths that the function generate_item produce.

    These paths are:
    - cloudprob_path_dest (.tif)
    - floodmap_path_dest. (.geojson)
    - gt_path (.tif)
    - meta_parent_path (.tif)
    - permanent_water_image_path_dest (.tif) or None if not permanent_water_available
    - s2_image_path_dest (.tif)

    Args:
        output_path: Path to produce the outputs
        file_name:
        permanent_water_available:
        clouds_available:
        mkdirs: make dirs if needed for the output paths

    Returns:
        cloudprob_path_dest, floodmap_path_dest, gt_path, meta_parent_path, permanent_water_image_path_dest, s2_image_path_dest

    """
    if permanent_water_available:
        permanent_water_image_path_dest = os.path.join(output_path, "PERMANENTWATERJRC", file_name+".tif").replace("\\", "/")
    else:
        permanent_water_image_path_dest = None

    output_path = str(output_path)
    s2_image_path_dest = os.path.join(output_path,"S2",file_name+".tif").replace("\\", "/")
    meta_parent_path = os.path.join(output_path,"meta",file_name+".json").replace("\\", "/")

    if clouds_available:
        cloudprob_path_dest = os.path.join(output_path, "cloudprob",file_name+".tif").replace("\\", "/")
    else:
        cloudprob_path_dest = None

    floodmap_path_dest = os.path.join(output_path,"floodmaps",file_name+".geojson").replace("\\", "/")
    gt_path = os.path.join(output_path,"gt",file_name+".tif").replace("\\", "/")

    # makedir if not gs
    if mkdirs and not s2_image_path_dest.startswith("gs"):
        fs = utils.get_filesystem(s2_image_path_dest)
        for f in [s2_image_path_dest, meta_parent_path, cloudprob_path_dest, floodmap_path_dest, gt_path, permanent_water_image_path_dest]:
            if f is not None:
                fs.makedirs(os.path.dirname(f), exist_ok=True)

    return cloudprob_path_dest, floodmap_path_dest, gt_path, meta_parent_path, permanent_water_image_path_dest, s2_image_path_dest



BANDS_S2_WITH_QA60 = ["B1", "B2", "B3", "B4", "B5",
            "B6", "B7", "B8", "B8A", "B9",
            "B10", "B11", "B12", "QA60"]
"""
Map from names in the floodmap to rasterised value of watermask (see function compute_water)

The meaning of the codes are: 
 {0: 'land', 1: 'flood', 2: 'hydro', 3: 'permanent_water_jrc'}

"""

CODES_FLOODMAP = {
    # CopernicusEMS (flood)
    'Flooded area': 1,
    'Previous flooded area': 1,
    'Not Applicable': 1,
    'Not Application': 1,
    'Flood trace': 1,
    'Dike breach': 1,
    'Standing water': 1,
    'Erosion': 1,
    'River': 2,
    'Riverine flood': 1,
    # CopernicusEMS (hydro)
    'BH140-River': 2,
    'BH090-Land Subject to Inundation': 0,
    'BH080-Lake': 2,
    'BA040-Open Water': 2,
    'BA030-Island': 0, # islands are excluded! see filter_land func
    'BH141-River Bank': 2,
    'BH170-Natural Spring': 2,
    'BH130-Reservoir': 2,
    'BH141-Stream': 2,
    'BA010-Coastline': 0,
    'BH180-Waterfall': 2,
    # UNOSAT
    "preflood water": 2,
    # "Flooded area": 1,  # 'flood water' DUPLICATED
    "flood-affected land / possible flood water": 1,
    # "Flood trace": 1,  # 'probable flash flood-affected land' DUPLICATED
    "satellite detected water": 1,
    # "Not Applicable": 1,  # unknown see document DUPLICATED
    "possible saturated, wet soil/ possible flood water": 1,
    "aquaculture (wet rice)": 1,
    "tsunami-affected land": 1,
    "ran of kutch water": 1,
    "maximum flood water extent (cumulative)": 1
}

CLASS_LAND_COPERNICUSEMSHYDRO = ["BH090-Land Subject to Inundation", "BA030-Island", 'BA010-Coastline']

ACCEPTED_FIELDS = list(CODES_FLOODMAP.keys()) + CLASS_LAND_COPERNICUSEMSHYDRO


# Unosat names definition https://docs.google.com/document/d/1i-Fz0o8isGTpRr39HqvUOQBs0yh8_Pz_WcyF5JK0bM0/edit#heading=h.3neqeg3hyp0t
UNOSAT_CLASS_TO_TXT = {
    0: "preflood water",
    1: "Flooded area",  # 'flood water'
    2: "flood-affected land / possible flood water",
    3: "Flood trace",  # 'probable flash flood-affected land'
    4: "satellite detected water",
    5: "Not Applicable",  # unknown see document
    6: "possible saturated, wet soil/ possible flood water",
    9: "aquaculture (wet rice)",
    14: "tsunami-affected land",
    77: "ran of kutch water",
    99: "maximum flood water extent (cumulative)"
}

RENAME_SATELLITE = {"landsat 5": "Landsat-5",
                    "landsat 8": "Landsat-8",
                    "landsat-8": "Landsat-8",
                    "landsat 7": "Landsat-7",
                    "sentinel-1": "Sentinel-1",
                    "pleadies": "Pleiades-1A-1B",
                    "sentinel-2": "Sentinel-2",
                    "radarsat-1": "RADARSAT-1",
                    "radarsat-2": "RADARSAT-2",
                    "terrasar-x": "TERRASAR-X",
                    'spot 6' : "SPOT-6-7",
                    "worldview-2": "WorldView-2"}

SATELLITE_TYPE = {'COSMO-SkyMed': "SAR",
                  'GeoEye-1': "SAR",
                  'Landsat-5': "Optical",
                  'Landsat-7': "Optical",
                  'Landsat-8': "Optical",
                  'PlanetScope': "Optical",
                  'Pleiades-1A-1B': "Optical",
                  'RADARSAT-1': "SAR",
                  'RADARSAT-2': "SAR",
                  'SPOT-6-7': "Optical",
                  'Sentinel-1': "SAR",
                  'Sentinel-2': "Optical",
                  'TERRASAR-X': "SAR",
                  'WorldView-1': "Optical",
                  'WorldView-2': "Optical",
                  'WorldView-3': "Optical",
                  'alos palsar': "SAR",
                  'asar imp': "SAR",
                  'dmc': "Optical",
                  'earth observing 1': "Optical",
                  'modis-aqua': "Optical",
                  'modis-terra': "Optical",
                  'spot4': "Optical"}


from typing import Tuple, Optional, List, Callable, Dict
from torch.utils.data import DataLoader
from ml4floods.data.worldfloods.dataset import WorldFloodsDatasetTiled, WorldFloodsDataset
import pytorch_lightning as pl
from ml4floods.preprocess.tiling import WindowSize
from ml4floods.preprocess.utils import get_list_of_window_slices


class WorldFloodsDataModule(pl.LightningDataModule):
    """A prepackaged WorldFloods Pytorch-Lightning data module
    This initializes a module given a set a directory with a subdirectory
    for the training and testing data ("image_folder" and "target_folder").
    Then we can search through the directory and load the images found. It
    creates the train, val and test datasets which then can be used to initialize
    the dataloaders. This is pytorch lightning compatible which can be used with
    the training fit framework.

    Args:
        input_folder (str): the input folder sub_directory
        target_folder (str): the target folder sub directory
        train_transformations (Callable): the transformations used within the
            training data module
        test_transformations (Callable): the transformations used within the
            testing data module
        window_size (Tuple[int,int]): the window size used to tile the images
            for training
        batch_size (int): the batchsize used for the dataloader
        bands (List(int)): the bands to be selected from the images

    Attributes:
        train_transform (Callable): the transformations used within the
            training data module
        test_transform (Callable): the transformations used within the
            testing data module
        bands (List(int)): the bands to be selected from the images
        image_prefix (str): the input folder sub_directory
        gt_prefix (str): the target folder sub directory
        window_size (Tuple[int,int]): the window size used to tile the images
            for training
        filter_windows (Callable): function to filter the training tiles by 
            number of invalid and cloud pixels 
        filenames_train_test (Dict): path to images and ground truth for 
            the training, validation and test splits 
      
    Example:
        >>> from ml4floods.data.worldfloods.lightning import WorldFloodsDataModule
        >>> wf_dm = WorldFloodsDataModule()
        >>> wf_dm.prepare_data()
        >>> wf_dm.setup()
        >>> train_dl = wf_dm.train_dataloader()
    """

    def __init__(
        self,
        filenames_train_test: Dict,
        input_folder: str = "S2",
        target_folder: str = "gt",
        train_transformations: Optional[Callable] = None,
        test_transformations: Optional[Callable] = None,
        window_size: Tuple[int, int] = (64, 64),
        batch_size: int = 32,
        bands: List[int] = [1, 2, 3],
        num_workers:int = 4,
        num_workers_val:int = 0,
        num_workers_test: int = 0,
        filter_windows:Callable = None,
        lock_read: bool = False,
    ):
        super().__init__()
        self.train_transform = train_transformations
        self.test_transform = test_transformations
        self.num_workers = num_workers
        self.num_workers_test = num_workers_test
        self.num_workers_val = num_workers_val
        self.lock_read = lock_read

        # self.dims is returned when you call dm.size()
        # Setting default dims here because we know them.
        # Could optionally be assigned dynamically in dm.setup()
        self.bands = bands
        self.batch_size = batch_size
        # Prefixes
        self.image_prefix = input_folder
        self.gt_prefix = target_folder
        self.filter_windows = filter_windows
        self.window_size = WindowSize(height=window_size[0], width=window_size[1])
        self.filenames_train_test = filenames_train_test

        files = {}
        splits = ["train", "test", "val"]

        # loop through the naming splits
        for isplit in splits:
                # TODO we might could use the train_test_split dict directly to avoid using image_prefix and gt_prefix
                files[isplit] = self.filenames_train_test[isplit][self.image_prefix]

        # save filenames
        self.train_files = files["train"]
        self.val_files = files["val"]
        self.test_files = files["test"]

    def prepare_data(self):
        """Does Nothing for now. Here for compatibility."""
        # TODO: here we can check for correspondence between the files
        pass

    def setup(self, stage=None):
        """This creates the PyTorch dataset given the preconfigured
        file paths.
        """

        self.train_dataset = WorldFloodsDatasetTiled(
            list_of_windows=get_list_of_window_slices(self.train_files, window_size=self.window_size),
            image_prefix=self.image_prefix,
            gt_prefix=self.gt_prefix,
            bands=self.bands,
            transforms=self.train_transform,
            lock_read=self.lock_read
        )
        if self.filter_windows is not None:
            self.train_dataset.list_of_windows = self.filter_windows(self.train_dataset)

        self.val_dataset = WorldFloodsDatasetTiled(
            list_of_windows=get_list_of_window_slices(
                self.val_files, window_size=self.window_size
            ),
            image_prefix=self.image_prefix,
            gt_prefix=self.gt_prefix,
            bands=self.bands,
            transforms=self.test_transform,
            lock_read=self.lock_read
        )

        self.test_dataset = WorldFloodsDataset(
            image_files=self.test_files,
            image_prefix=self.image_prefix,
            gt_prefix=self.gt_prefix,
            bands=self.bands,
            transforms=self.test_transform,
            lock_read=self.lock_read
        )

    def train_dataloader(self):
        """Initializes and returns the training dataloader"""
        return DataLoader(self.train_dataset, batch_size=self.batch_size,
                          num_workers=self.num_workers, shuffle=True)

    def val_dataloader(self, num_workers=None):
        """Initializes and returns the validation dataloader"""
        num_workers = num_workers or self.num_workers_val
        return DataLoader(self.val_dataset, batch_size=self.batch_size,
                          num_workers=num_workers, shuffle=False)

    def test_dataloader(self, num_workers=None):
        """Initializes and returns the test dataloader"""
        num_workers = num_workers or self.num_workers_test
        return DataLoader(self.test_dataset, batch_size=1,
                          num_workers=num_workers, shuffle=False)



import atexit
import logging
import os
import subprocess
import time
from signal import SIGTERM

import requests

logging.basicConfig(level=logging.INFO)

class TileServer:
    
    def __init__(self):
        self.port=None
        self.server_pid=None
        self.logger=logging.getLogger('TILESERVER')
    
    def serve(self, workers, port):
        atexit.register(self.stop) # make sure the tileserver is stopped if this object dies
        P = subprocess.Popen(['gunicorn', f'--workers={workers}','ml4floods.serve.tileserver.app:app', f'--bind=0.0.0.0:{port}'])
        self.server_pid = P.pid
        self.port=port
        time.sleep(3)  # sleep for 3s to let workers boot
        self.logger.info(f'Started with pid {self.server_pid} port {self.port}')
        
    def stop(self):
        if self.port:
            self.logger.info('Stopping server')
            #r = requests.get('http://localhost:'+str(self.port)+'/terminate')
            os.kill(self.server_pid, SIGTERM)
            self.port=None # turn the port off again
            self.server_pid = None
            time.sleep(3) # sleep for 3s to let working shutdown
            self.logger.info('Server stopped.')
        else:
            self.logger.info('No server running!')
            
            
class ModelServer:
    
    def __init__(self):
        self.port=None
        self.server_pid=None
        self.logger=logging.getLogger('MODELSERVER')
        
    def serve(self, workers, port):
        atexit.register(self.stop) # make sure the tileserver is stopped if this object dies
        P = subprocess.Popen(['gunicorn', f'--workers={workers}','ml4floods.serve.modelserver.app:app', f'--bind=127.0.0.1:{port}'])
        self.server_pid = P.pid
        self.port=port
        time.sleep(3)  # sleep for 3s to let workers boot
        self.logger.info(f'Started with pid {self.server_pid} port {self.port}')
        
    def stop(self):
        if self.port:
            self.logger.info('Stopping server')
            #r = requests.get('http://localhost:'+str(self.port)+'/terminate')
            os.kill(self.server_pid, SIGTERM)
            self.port=None # turn the port off again
            self.server_pid = None
            time.sleep(3) # sleep for 3s to let working shutdown
            self.logger.info('Server stopped.')
        else:
            self.logger.info('No server running!')
            
        
    



import json
import logging
import math
import os
from datetime import datetime

from dateutil import parser
from fs import open_fs

from ml4floods.data.index.geographic_index import GeographicIndex
from ml4floods.data.index.indexer_arg_parser import IndexerArgParser

LOGGER = logging.getLogger(__name__)

MASTER_PREFIX = "worldfloods/tiffimages"
METADATA_PREFIX = "meta"
PROVIDER_GEOTIFF_PREFIXES = ["S2", "L8", "gt"]
CLOUDMASK_PREFIX = "cloudprob"
FLOODMAP_PREFIX = "floodmaps"

def file_exists(path):
    with open_fs(path+"?strict=False") as fs:
        return fs.exists('')

def index_worldfloods(fs_prefix: str, out_path: str) -> None:
    """Function to index all the files within the fs_prefix path.

    Args:
        fs_prefix (str): Source directory for the data to be indexed.
        out_path (str): Destination directory to store the indexed data.
    """
    idx = GeographicIndex()
    count = 0
    totalcount = 0
    metadata_timestamp = datetime.utcnow()
    print("Enumerating metadata...")
    print("Note: This is practically instant with local block storage, but may take several minutes if running from a cloud storage bucket (GCP, AWS, Azure, etc.)")
    with open_fs(fs_prefix + os.path.join(MASTER_PREFIX, METADATA_PREFIX)+"?strict=False") as fs:
        for file in fs.listdir(''):
            if file.endswith(".json"):
                totalcount += 1
                with open_fs(fs_prefix + os.path.join(MASTER_PREFIX, METADATA_PREFIX, file)+"?strict=False", "r") as fs2:
                    with fs2.open('', "r") as f: 
                        metadata = json.load(f)

                # Some metadata is corrupted -- full of NaN. Let's not even try to deal with those records for now.
                # These records also tend not to have bounds
                if "bounds" in metadata:
                    bounds = metadata["bounds"]
                    satdate = satdate = parser.parse(metadata["satellite date"])
                    s2satdate = satdate
        
                    # Gather up the relevant files and some basic time/date metadata from them
                    relevant_files = []
                    basename = os.path.splitext(file)[0]
        
                    if "s2metadata" in metadata and len(metadata["s2metadata"]) == 1:
                        s2satdate = satdate = parser.parse(metadata["s2metadata"][0]["date_string"])
        
                    if file_exists(os.path.join(fs_prefix, MASTER_PREFIX, FLOODMAP_PREFIX, basename+".shp")):
                        relevant_files.append({"type": "floodmap", "last_modified": satdate, "path": os.path.join(MASTER_PREFIX, FLOODMAP_PREFIX, basename+".shp")})
        
                    if file_exists(os.path.join(fs_prefix, MASTER_PREFIX, CLOUDMASK_PREFIX, basename+".tif")):
                        relevant_files.append({"type": "cloudmask", "last_modified": satdate, "path": os.path.join(MASTER_PREFIX, CLOUDMASK_PREFIX, basename+".tif")})
        
                    count += 1
                    for sat_prefix in PROVIDER_GEOTIFF_PREFIXES:
                        sat_path = os.path.join(fs_prefix, MASTER_PREFIX, sat_prefix, basename+".tif")
                        if file_exists(sat_path):
                            relevant_files.append({"type": "satellite_image", "last_modified": s2satdate if sat_prefix == "S2" else satdate, "provider_id": sat_prefix, "path": os.path.join(MASTER_PREFIX, sat_prefix, basename+".tif")})

                    relevant_files.append({"type": "meta", "last_modified": metadata_timestamp, "path": os.path.join(MASTER_PREFIX, METADATA_PREFIX, basename+".json")})
        
                    # Add these relevant files to each 1-degree x 1-degree bin
                    # Covered area represented in half-open intervals:
                    # [min_lat, max_lat)
                    # [min_lon, max_lon)
                    min_lat = math.floor(bounds[1])
                    min_lon = math.floor(bounds[0])
                    max_lat = math.floor(bounds[3]+1)
                    max_lon = math.floor(bounds[2]+1)
                    for lat in range(min_lat, max_lat):
                        for lon in range(min_lon, max_lon):
                            idx.append_at_coords(lat, lon, relevant_files)
    
    LOGGER.info(f"{count} valid records indexed out of {totalcount}")
    
    idx.save_index(out_path)

if __name__ == "__main__":
    args = IndexerArgParser().parse_args()

    numeric_level = getattr(logging, args.log_level.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError('Invalid log level: %s' % loglevel)
    logging.basicConfig(level=numeric_level)

    index_worldfloods(args.worldfloods_path, args.output_path)



import logging
import os
from typing import Dict, Optional, Tuple, List, Union

import geopandas as gpd
import numpy as np
import rasterio
import rasterio.windows
from rasterio import features

from ml4floods.data.worldfloods.configs import BANDS_S2, BANDS_L8
from ml4floods.data.config import CODES_FLOODMAP
from ml4floods.data import utils
from skimage.morphology import binary_opening, disk
import torch


def compute_water(
    tiffs2: str,
    floodmap: gpd.GeoDataFrame,
    window: Optional[rasterio.windows.Window] = None,
    permanent_water_path: str = None,
    keep_streams: bool = False,
) -> np.ndarray:
    """
    Rasterise flood map and add JRC permanent water layer

    Args:
        tiffs2: Tif file S2 (either remote or local)
        floodmap: geopandas dataframe with the annotated polygons
        window: rasterio.windows.Window to read. Could also be slices (slice(100, 200), slice(100, 200)
        permanent_water_path: Whether or not user JRC permanent water layer (from tifffimages/PERMANENTWATERJRC folder)
        keep_streams: A boolean flag to indicate whether to include streams in the water mask

    Returns:
        water_mask : np.int16 raster same shape as tiffs2 {-1: invalid, 0: land, 1: flood, 2: hydro, 3: permanentwaterjrc}
    """

    with utils.rasterio_open_read(tiffs2) as src_s2:
        if window is None:
            out_shape = src_s2.shape
            transform = src_s2.transform
        else:
            out_shape = rasterio.windows.shape(
                window, height=src_s2.height, width=src_s2.width
            )
            transform = rasterio.windows.transform(window, src_s2.transform)
        target_crs = str(src_s2.crs).lower()

    if str(floodmap.crs).lower() != target_crs:
        floodmap.to_crs(crs=target_crs, inplace=True)

    # area_of_interest contains the extent that was labeled (values out of this pol should be marked as invalid)
    floodmap_aoi = floodmap[floodmap["w_class"] == "area_of_interest"]

    floodmap_rasterise = floodmap
    if floodmap_aoi.shape[0] > 0:
        floodmap_rasterise = floodmap_rasterise[floodmap["w_class"] != "area_of_interest"]

    if not keep_streams:
        floodmap_rasterise = floodmap_rasterise[floodmap["source"] != "hydro_l"]

    shapes_rasterise = (
        (g, CODES_FLOODMAP[w])
        for g, w in floodmap_rasterise[["geometry", "w_class"]].itertuples(
            index=False, name=None
        ) if g and not g.is_empty
    )

    water_mask = features.rasterize(
        shapes=shapes_rasterise,
        fill=0,
        out_shape=out_shape,
        dtype=np.int16,
        transform=transform,
        all_touched=keep_streams,
    )

    # Load valid mask using the area_of_interest polygons (valid pixels are those within area_of_interest polygons)
    if floodmap_aoi.shape[0] > 0:
        shapes_rasterise = (
            (g, 1)
            for g, w in floodmap_aoi[["geometry", "w_class"]].itertuples(
                index=False, name=None
            ) if g and not g.is_empty
        )
        valid_mask = features.rasterize(
            shapes=shapes_rasterise,
            fill=0,
            out_shape=out_shape,
            dtype=np.uint8,
            transform=transform,
            all_touched=True,
        )
        water_mask[valid_mask == 0] = -1

    if permanent_water_path is not None:
        logging.info("\t Adding permanent water")
        with utils.rasterio_open_read(permanent_water_path) as srcperm:
            permanent_water = srcperm.read(1, window=window)

        # Set to permanent water
        # Only interested in permanent water labelled as 3 and valid water masks.
        # Adding the third label for permanent water in water mask.
        water_mask[(water_mask != -1) & (permanent_water == 3)] = 3

        # Seasonal water (permanent_water == 2) will not be used

    return water_mask


def read_s2img_cloudmask(
    s2tiff: str,
    window: Optional[rasterio.windows.Window] = None,
    cloudprob_image_path: Optional[str] = None,
    bands_read: Optional[List[int]] = None
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Helper function to load a s2 image and its cloud mask

    Args:
        s2tiff: path to S2 tiff file
        window: Optional window to read
        cloudprob_image_path: Optional path to geojson or tiff file with cloud mask.
        bands_read: bands to read from `BANDS_S2` [0-based]

    Returns:
        s2img: (C, H, W) array with len(bands_read) channels.
        cloud_prob: (H, W) array with cloud probability

    """
    if bands_read is None:
        bands_read = list(range(1, len(BANDS_S2) + 1))
    else:
        bands_read = [b+1 for b in bands_read]

    with utils.rasterio_open_read(s2tiff) as s2_rst:
        s2_img = s2_rst.read(bands_read, window=window)
        bands = s2_rst.descriptions
        cloudprob_in_lastband = (len(bands) > 14) and (bands[14] == "probability")
        transform = s2_rst.transform if window is None else rasterio.windows.transform(window, s2_rst.transform)
        crs = str(s2_rst.crs).lower()
        shape = s2_img.shape[1:]

    if cloudprob_image_path:
        fs = utils.get_filesystem(cloudprob_image_path)
    else:
        fs = None

    if fs and fs.exists(cloudprob_image_path):
        ext = os.path.splitext(cloudprob_image_path)[1]
        if ext == ".tif":
            with utils.rasterio_open_read(cloudprob_image_path) as cld_rst:
                cloud_prob = cld_rst.read(1, window=window)

        elif  ext == ".geojson":
            # check if empty file -> cloud_masks=np.zeros(shape)
            clouds_vec = utils.read_geojson_from_gcp(cloudprob_image_path)
            if clouds_vec.shape[0] == 0:
                cloud_prob = np.zeros(s2_img.shape[1:], dtype=np.float32)
            else:
                # Convert clouds_vec crs to s2 crs
                if str(clouds_vec.crs).lower() != crs:
                    clouds_vec.to_crs(crs=crs, inplace=True)

                clouds_vec = clouds_vec[clouds_vec["class"] == "CLOUD"]
                shapes_rasterise = (
                    (g, 1)
                    for g,w in clouds_vec[["geometry", "class"]].itertuples(
                    index=False, name=None
                )
                )
                cloud_prob = features.rasterize(
                    shapes=shapes_rasterise,
                    fill=0,
                    out_shape=shape,
                    dtype=np.float32,
                    transform=transform,
                    all_touched=True,
                )
    elif cloudprob_in_lastband:
        with utils.rasterio_open_read(s2tiff) as s2_rst:
            last_band = s2_rst.count
            cloud_prob = s2_rst.read(last_band, window=window)
            cloud_prob = (
                cloud_prob.astype(np.float32) / 100.0
            )  # cloud mask in the last band is from 0 - 100
    else:
        from ml4floods.data import cloud_masks
        # Compute cloud mask
        cloud_prob = cloud_masks.compute_cloud_mask(s2_img)

    return s2_img, cloud_prob


def generate_land_water_cloud_gt(
    s2_image_path: str,
    floodmap: gpd.GeoDataFrame,
    metadata_floodmap: Dict,
    window: Optional[rasterio.windows.Window] = None,
    keep_streams: bool = False,
    permanent_water_image_path: Optional[str] = None,
    cloudprob_image_path: Optional[str] = None,
) -> Tuple[np.ndarray, Dict]:
    """
    Old ground truth generating function (inherited from worldfloods_internal.compute_meta_tif.generate_mask_meta_clouds)

    Args:
        s2_image_path:
        floodmap:
        metadata_floodmap:
        window:
        permanent_water_image_path:
        keep_streams: A boolean flag to indicate whether to include streams in the water mask
        cloudprob_image_path:

    Returns:
        gt (np.ndarray): (H, W) np.uint8 array with encodding: {0: "invalid", 1: "land", 2: "water", 3: "cloud"}
        meta: dictionary with metadata information

    """
    # open floodmap with geopandas
    # =========================================
    # Generate Cloud Mask given S2 Data
    # =========================================
    s2_img, cloud_mask = read_s2img_cloudmask(
        s2_image_path,
        window=window,
        cloudprob_image_path=cloudprob_image_path,
    )
    # =========================================
    # Compute Water Mask
    # =========================================
    water_mask = compute_water(
        s2_image_path,
        floodmap,
        window=window,
        permanent_water_path=permanent_water_image_path,
        keep_streams=keep_streams,
    )

    gt = _generate_gt_v1_fromarray(s2_img, cloudprob=cloud_mask, water_mask=water_mask)

    # Compute metadata of the ground truth
    metadata = metadata_floodmap.copy()
    metadata["gtversion"] = "v1"
    metadata["encoding_values"] = {0: "invalid", 1: "land", 2: "water", 3: "cloud"}
    metadata["shape"] = list(water_mask.shape)
    metadata["s2_image_path"] = os.path.basename(s2_image_path)
    metadata["permanent_water_image_path"] = (
        os.path.basename(permanent_water_image_path)
        if permanent_water_image_path is not None
        else "None"
    )
    metadata["cloudprob_image_path"] = (
        os.path.basename(cloudprob_image_path)
        if cloudprob_image_path is not None
        else "None"
    )
    metadata["method clouds"] = "s2cloudless"

    # Compute stats of the GT
    metadata["pixels invalid S2"] = int(np.sum(gt == 0))
    metadata["pixels clouds S2"] = int(np.sum(gt == 3))
    metadata["pixels water S2"] = int(np.sum(gt == 2))
    metadata["pixels land S2"] = int(np.sum(gt == 1))

    # Compute stats of the water mask
    metadata["pixels flood water S2"] = int(np.sum(water_mask == 1))
    metadata["pixels hydro water S2"] = int(np.sum(water_mask == 2))
    metadata["pixels permanent water S2"] = int(np.sum(water_mask == 3))

    # Sanity check: values of masks add up (water_mask is set to zero in _generate_gt_v1_fromarray where clouds or invalids)
    assert (
        metadata["pixels flood water S2"]
        + metadata["pixels hydro water S2"]
        + metadata["pixels permanent water S2"]
    ) == metadata[
        "pixels water S2"
    ], f'Different number of water pixels than expected {metadata["pixels flood water S2"]} {metadata["pixels hydro water S2"]} {metadata["pixels permanent water S2"]}, {metadata["pixels water S2"]} '

    with utils.rasterio_open_read(s2_image_path) as s2_src:
        metadata["bounds"] = s2_src.bounds
        metadata["crs"] = s2_src.crs
        metadata["transform"] = s2_src.transform

    return gt, metadata


def generate_water_cloud_binary_gt(
    s2_image_path: str,
    floodmap: gpd.GeoDataFrame,
    metadata_floodmap: Dict,
    window: Optional[rasterio.windows.Window] = None,
    keep_streams: bool = False,
    permanent_water_image_path: Optional[str] = None,
    cloudprob_image_path: Optional[str] = None,
) -> Tuple[np.ndarray, Dict]:
    """
    New ground truth generating function for multioutput binary classification

    Args:
        s2_image_path:
        floodmap:
        metadata_floodmap: Metadata of the floodmap (if satellite is optical will mask the land/water GT)
        window:
        keep_streams:
        permanent_water_image_path:
        cloudprob_image_path:

    Returns:
        gt (np.ndarray): (2, H, W) np.uint8 array where:
            First channel encodes the cloud GT {0: invalid, 1: clear, 2: cloud}
            Second channel encodes the land/water GT {0: invalid, 1: land, 2: water}
        meta: dictionary with metadata information

    """

    # =========================================
    # Generate Cloud Mask given S2 Data
    # =========================================
    s2_img, cloud_mask = read_s2img_cloudmask(
        s2_image_path,
        window=window,
        cloudprob_image_path=cloudprob_image_path,
    )
    custom_cloud_mask = (cloudprob_image_path is not None) and cloudprob_image_path.endswith(".geojson")

    water_mask = compute_water(
        s2_image_path,
        floodmap,
        window=window,
        permanent_water_path=permanent_water_image_path,
        keep_streams=keep_streams,
    )

    # # TODO this should be invalid if it is Sentinel-2 and it is exactly the same date ('satellite date' is the same as the date of retrieval of s2tiff)
    # if metadata_floodmap["satellite"] == "Sentinel-2":
    #     invalid_clouds_threshold = 0.5
    # else:
    #     invalid_clouds_threshold = None

    gt = _generate_gt_fromarray(
        s2_img,
        cloudprob=cloud_mask,
        water_mask=water_mask,
        custom_clouds=custom_cloud_mask,
    )

    # Compute metadata of the ground truth
    metadata = metadata_floodmap.copy()

    metadata["gtversion"] = "v2"
    metadata["encoding_values"] = [
        {0: "invalid", 1: "clear", 2: "cloud"},
        {0: "invalid", 1: "land", 2: "water"},
    ]
    metadata["shape"] = list(water_mask.shape)
    metadata["s2_image_path"] = os.path.basename(s2_image_path)
    metadata["permanent_water_image_path"] = (
        os.path.basename(permanent_water_image_path)
        if permanent_water_image_path is not None
        else "None"
    )
    metadata["cloudprob_image_path"] = (
        os.path.basename(cloudprob_image_path)
        if cloudprob_image_path is not None
        else "None"
    )
    metadata["method clouds"] = "s2cloudless"

    # Compute stats of the GT
    metadata["pixels invalid S2"] = int(np.sum(gt[0] == 0))
    metadata["pixels clouds S2"] = int(np.sum(gt[0] == 2))
    metadata["pixels water S2"] = int(np.sum(gt[1] == 2))
    metadata["pixels land S2"] = int(np.sum(gt[1] == 1))

    # Compute stats of the water mask
    metadata["pixels flood water S2"] = int(np.sum(water_mask == 1))
    metadata["pixels hydro water S2"] = int(np.sum(water_mask == 2))
    metadata["pixels permanent water S2"] = int(np.sum(water_mask == 3))

    with utils.rasterio_open_read(s2_image_path) as s2_src:
        metadata["bounds"] = s2_src.bounds
        metadata["crs"] = s2_src.crs
        metadata["transform"] = s2_src.transform

    return gt, metadata


def _generate_gt_v1_fromarray(
    s2_img: np.ndarray, cloudprob: np.ndarray, water_mask: np.ndarray
) -> np.ndarray:
    """
    Generate Ground Truth of V1 of WorldFloods multi-class classification problem

    :param s2_img: (C, H, W) used to find the invalid values in the input
    :param cloudprob:  probability value between [0,1]
    :param water_mask:  {-1: invalid, 0: land, 1: flood, 2: hydro, 3: permanentwaterjrc}
    :return: (H, W) np.uint8 {0: invalid, 1: land, 2:water, 3: cloud}
    """

    invalids = np.all(s2_img == 0, axis=0) | (water_mask == -1)

    # Set cloudprobs to zero in invalid pixels
    cloudprob[invalids] = 0
    cloudmask = cloudprob > 0.5

    # Set watermask values for compute stats
    water_mask[invalids | cloudmask] = 0

    # Create gt mask {0: invalid, 1:land, 2: water, 3: cloud}
    gt = np.ones(water_mask.shape, dtype=np.uint8)
    gt[water_mask > 0] = 2
    gt[cloudmask] = 3
    gt[invalids] = 0

    return gt

def get_brightness(s2_image:Union[np.ndarray, torch.Tensor], channels_input:Optional[List[int]]=None,
                   collection_name:str="S2") -> Union[np.ndarray, torch.Tensor]:
    """
    Returns brightness of visible bands of s2
    Args:
        s2_image: (C, H, W) array
        channels_input: 0-based list of indexes of s2_image channels (expected that len(channels_input) == s2_image.shape[0])
        collection_name: S2 or Landsat
    Returns:
        (H, W) array
    """
    if channels_input is not None:
        assert len(channels_input) == s2_image.shape[0], \
            f"Given {len(channels_input)} channels expected {s2_image.shape[0]}"

        if collection_name == "Landsat":
            bands_read_names = [BANDS_L8[iband] for iband in channels_input]
        else:
            bands_read_names = [BANDS_S2[i] for i in channels_input]
    else:
        assert len(BANDS_S2) == s2_image.shape[0], \
            f"Given {len(BANDS_S2)} channels expected {s2_image.shape[0]}"
        if collection_name == "Landsat":
            bands_read_names = BANDS_L8
        else:
            bands_read_names = BANDS_S2

    idxs = [bands_read_names.index(b) for b in ["B4", "B3", "B2"]]

    if isinstance(s2_image, torch.Tensor):
        rgb_img = s2_image[idxs, ...].float() ** 2
        return torch.sqrt(torch.sum(rgb_img, dim=0))

    rgb_img = s2_image[idxs, ...].astype(np.float32) ** 2
    return np.sqrt(np.sum(rgb_img, axis=0))


CLOUDS_THRESHOLD = .5
BRIGHTNESS_THRESHOLD = 3_500



def _generate_gt_fromarray(
    s2_img: np.ndarray,
    cloudprob: np.ndarray,
    water_mask: np.ndarray,
    custom_clouds: bool=False,
) -> np.ndarray:
    """

    Generate Ground Truth of WorldFloods Extra (multi-output binary classification problem)

    Args:
        s2_img: (C, H, W) array
        cloudprob: (H, W) array
        water_mask: (H, W) array {-1: invalid, 0: land, 1: flood, 2: hydro, 3: permanentwaterjrc}
        custom_clouds: whether it uses a custom gt or not.
            If it is not custom it will mask the bright cloud pixels in the s2image in the water mask.

    Returns:
        (2, H, W) np.uint8 array where:
            First channel encodes {0: invalid, 1: clear, 2: cloud}
            Second channel encodes {0: invalid, 1: land, 2: water}

        A pixel is set to invalid if it's invalid in the water_mask layer or invalid in the s2_img (all values to zero)

    """

    invalids = np.any(np.isnan(s2_img), axis=0) | np.all(s2_img[:len(BANDS_S2)] == 0, axis=0) | (water_mask == -1)

    # Set cloudprobs to zero in invalid pixels
    cloudgt = np.ones(water_mask.shape, dtype=np.uint8)
    cloudgt[cloudprob > CLOUDS_THRESHOLD] = 2
    cloudgt[invalids] = 0

    # For clouds we could set to invalid only if the s2_img is invalid (not the water mask)?

    # Set watermask values for compute stats
    watergt = np.ones(water_mask.shape, dtype=np.uint8)  # whole image is 1
    watergt[water_mask >= 1] = 2  # only water is 2
    watergt[invalids] = 0

    clouds = cloudprob > CLOUDS_THRESHOLD
    if custom_clouds:
        # set to invalid in watergt clouds in gt (assume manually added to include only thick clouds!)
        watergt[clouds] = 0
    else:
        # set to invalid in watergt for bright clouds
        brightness = get_brightness(s2_img) # (H, W)
        clouds &= (brightness >= BRIGHTNESS_THRESHOLD)

        # binary opening of bright clouds
        open_clouds = binary_opening(clouds, disk(3)).astype(np.bool)

        # Set to invalid bright clouds
        watergt[open_clouds] = 0

    # if invalid_clouds_threshold is not None:
    #     # Set to invalid land pixels that are cloudy if the satellite is Sentinel-2
    #     watergt[(water_mask == 0) & (cloudprob > invalid_clouds_threshold)] = 0

    stacked_cloud_water_mask = np.stack([cloudgt, watergt], axis=0)

    return stacked_cloud_water_mask



from collections import namedtuple
from itertools import product
from pathlib import Path
from typing import Callable, List, Optional, Dict

import rasterio
from rasterio import windows
from rasterio.io import DatasetReader
import json

WindowSize = namedtuple("WindowSize", ["height", "width"])
WindowSlices = namedtuple("WindowSlices", ["file_name", "window"])


def load_windows(filename:str) -> List[WindowSlices]:
    with open(filename, "r") as fh:
        list_of_windows = [Dict_to_WindowSlices(dictio) for dictio in json.load(fh)["slices"]]
    return list_of_windows

def save_windows(list_windows:List[WindowSlices], filename:str) -> None:
    list_save = [WindowSlices_to_Dict(ws) for ws in list_windows]
    with open(filename, "w") as fh:
        json.dump({"slices": list_save}, fh)


def WindowSlices_to_Dict(ws: WindowSlices) -> Dict:
    return {
        "file_name" : ws.file_name,
        "window": {
            "col_off" : ws.window.col_off,
            "row_off": ws.window.row_off,
            "width": ws.window.width,
            "height": ws.window.height,
        }
    }

def Dict_to_WindowSlices(ds: Dict) -> WindowSlices:
    return WindowSlices(file_name=ds["file_name"],
                        window=windows.Window(col_off=ds["window"]["col_off"],
                                              row_off=ds["window"]["row_off"],
                                              width=ds["window"]["width"],
                                              height=ds["window"]["height"]))



def yield_window_tiles(
    ds: rasterio.io.DatasetReader, height: int = 128, width: int = 128, **kwargs
) -> rasterio.windows.Window:
    """a generator for rasterio specific slices given a rasterio dataset

    Args:
        ds (rasterio.io.DatasetReader): a rasterio dataset object
        height (int): the height for the slice
        width (int): the width for the slice

    Yields:
        window (rasterio.windows.Window): slicing
    """
    # extract the row height from the dataset
    n_columns, n_rows = ds.meta["width"], ds.meta["height"]

    # create the offsets
    offsets = product(range(0, n_columns, width), range(0, n_rows, height))

    for col_offset, row_offset in offsets:
        window = windows.Window(
            col_off=col_offset, row_off=row_offset, width=width, height=height, **kwargs
        )
        yield window


def get_window_tiles(
    ds: rasterio.io.DatasetReader, height: int = 128, width: int = 128, **kwargs
) -> List[rasterio.windows.Window]:
    """a generator for rasterio specific slices given a rasterio dataset

    Args:
        ds (rasterio.io.DatasetReader): a rasterio dataset object
        height (int): the height for the slice
        width (int): the width for the slice

    Yields:
        window (rasterio.windows.Window): slicing
    """
    # extract the row height from the dataset
    n_columns, n_rows = ds.meta["width"], ds.meta["height"]

    # create the offsets
    offsets = product(range(0, n_columns, width), range(0, n_rows, height))
    list_of_windows = []
    for col_offset, row_offset in offsets:
        iwindow = windows.Window(
            col_off=col_offset, row_off=row_offset, width=width, height=height, **kwargs
        )
        list_of_windows.append(iwindow)

    return list_of_windows


def save_tiles(
    file_name: str,
    dest_dir: str,
    bands: List[str],
    window_size: WindowSize,
    verbose: bool = False,
    n_samples: Optional[int] = None,
) -> None:
    """does tiling from a predefined window size and saves them to a directory

    Args:
        file_name (str): the filename to be opened by rasterio
        dest_dir (str): the directory where the data will be saved
        bands (List[str]): a list of bands to be accessed and saved
        window_size (WindowSize): a namedtuple with the height and width
        verbose (bool): flag to allow one to view the files saved

    Returns:
        None
    """
    with rasterio.open(file_name) as dataset:

        itile = 0

        # copy the metadata
        window_meta = dataset.meta.copy()

        for window_tile in get_window_tiles(
            dataset, window_size.height, window_size.width
        ):

            # copy the meta data
            window_meta["width"] = window_tile.width
            window_meta["height"] = window_tile.height
            window_meta["channels"] = bands
            window_meta["transform"] = windows.transform(window_tile, dataset.transform)

            # open the dataset with only the window selected
            sub_image = dataset.read(indexes=bands, window=window_tile)

            # create unique filename for the tile
            window_file_name = (
                f"{str(Path(file_name).stem)}_tile_{itile}{str(Path(file_name).suffix)}"
            )

            # filepath for saving
            output_tile_file_name = str(Path(dest_dir).joinpath(window_file_name))

            # open file and also save meta data
            with rasterio.open(output_tile_file_name, "w", **window_meta) as out_f:
                out_f.write(
                    dataset.read(window=window_tile, boundless=True, fill_value=0)
                )

            itile += 1
            if verbose:
                print(f"Saved: {window_file_name}")

            if n_samples is not None and n_samples == itile:
                break
    return None
